import itertools as it
import os
import random

from Bio.Phylo.TreeConstruction import BaseTree
import alifedata_phyloinformatics_convert as apc
import dendropy as dp
import networkx as nx
import pandas as pd
from phyloframe import legacy as pfl
import pytest
from tqdm import tqdm

from hstrat import hstrat
from hstrat._auxiliary_lib import generate_n, random_tree, seed_random
from hstrat.phylogenetic_inference.tree._impl import (
    build_tree_searchtable_python,
)

from .. import _impl as impl

assets_path = os.path.join(os.path.dirname(__file__), "..", "assets")


def test_empty_population():
    population = []
    tree = build_tree_searchtable_python(
        population,
    )

    assert len(tree) == 0
    assert pfl.alifestd_validate(tree)
    assert pfl.alifestd_is_chronologically_ordered(tree)


def test_dual_population_no_mrca():
    organism1 = hstrat.HereditaryStratigraphicColumn().CloneNthDescendant(100)
    organism2 = hstrat.HereditaryStratigraphicColumn().CloneNthDescendant(100)

    population = [organism1, organism2]
    names = ["foo", "bar"]

    with pytest.raises(ValueError):
        tree = build_tree_searchtable_python(
            population,
            taxon_labels=names,
        )

    tree = build_tree_searchtable_python(
        population,
        taxon_labels=names,
        force_common_ancestry=True,
    )
    tree["name"] = tree["taxon_label"]
    assert not pfl.alifestd_has_multiple_roots(tree)
    assert pfl.alifestd_validate(tree)
    assert pfl.alifestd_is_chronologically_ordered(tree)

    root_clade = BaseTree.Clade(name="Inner1")
    root_clade.clades = [
        BaseTree.Clade(branch_length=100.0, name="bar"),
        BaseTree.Clade(branch_length=100.0, name="foo"),
    ]
    true_tree = BaseTree.Tree(rooted=False, root=root_clade)
    assert (
        impl.tree_unweighted_robinson_foulds_distance(
            apc.alife_dataframe_to_biopython_tree(
                pfl.alifestd_collapse_unifurcations(tree),
                setup_branch_lengths=True,
            ),
            true_tree,
        )
        == 0.0
    )


@pytest.mark.parametrize(
    "orig_tree",
    [
        impl.setup_dendropy_tree(f"{assets_path}/grandchild_and_aunt.newick"),
        impl.setup_dendropy_tree(
            f"{assets_path}/grandchild_and_auntuncle.newick"
        ),
        impl.setup_dendropy_tree(f"{assets_path}/grandchild.newick"),
        impl.setup_dendropy_tree(
            f"{assets_path}/grandtriplets_and_aunt.newick"
        ),
        impl.setup_dendropy_tree(
            f"{assets_path}/grandtriplets_and_auntuncle.newick"
        ),
        impl.setup_dendropy_tree(f"{assets_path}/grandtriplets.newick"),
        impl.setup_dendropy_tree(f"{assets_path}/grandtwins_and_aunt.newick"),
        impl.setup_dendropy_tree(
            f"{assets_path}/grandtwins_and_auntuncle.newick"
        ),
        impl.setup_dendropy_tree(f"{assets_path}/grandtwins.newick"),
        impl.setup_dendropy_tree(f"{assets_path}/justroot.newick"),
        impl.setup_dendropy_tree(f"{assets_path}/triplets.newick"),
        impl.setup_dendropy_tree(f"{assets_path}/twins.newick"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
@pytest.mark.parametrize("wrap", [lambda x: x, hstrat.col_to_specimen])
def test_handwritten_trees(orig_tree, retention_policy, wrap):
    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_instrument=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(10),
    )

    reconst_df = build_tree_searchtable_python(
        [*map(wrap, extant_population)],
    )

    assert pfl.alifestd_validate(reconst_df)
    assert pfl.alifestd_is_chronologically_ordered(reconst_df)
    reconst_tree = apc.alife_dataframe_to_dendropy_tree(
        reconst_df,
        setup_edge_lengths=True,
    )
    reconst_tree.collapse_unweighted_edges()

    common_namespace = dp.TaxonNamespace()
    orig_tree.migrate_taxon_namespace(common_namespace)
    reconst_tree.migrate_taxon_namespace(common_namespace)

    original_distance_matrix = orig_tree.phylogenetic_distance_matrix()
    reconstructed_distance_matrix = reconst_tree.phylogenetic_distance_matrix()

    taxa = [node.taxon for node in orig_tree.leaf_node_iter()]

    for a, b in it.combinations(taxa, 2):
        assert original_distance_matrix.distance(
            a, b
        ) == reconstructed_distance_matrix.distance(a, b)


@pytest.mark.parametrize(
    "orig_tree",
    [
        pytest.param(
            impl.setup_dendropy_tree(f"{assets_path}/nk_ecoeaselection.csv"),
            marks=pytest.mark.heavy,
        ),
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
        impl.setup_dendropy_tree(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.recency_proportional_resolution_algo.Policy(3),
        hstrat.fixed_resolution_algo.Policy(5),
    ],
)
def test_reconstructed_mrca(orig_tree, retention_policy):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_instrument=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )

    reconst_df = build_tree_searchtable_python(
        extant_population,
    )
    assert "origin_time" in reconst_df

    assert pfl.alifestd_validate(reconst_df)
    assert pfl.alifestd_is_chronologically_ordered(reconst_df)
    reconst_tree = apc.alife_dataframe_to_dendropy_tree(
        reconst_df,
        setup_edge_lengths=True,
    )
    pdm = reconst_tree.phylogenetic_distance_matrix()

    assert len(list(reconst_tree.leaf_node_iter())) == len(extant_population)
    sorted_leaf_nodes = sorted(
        reconst_tree.leaf_node_iter(), key=lambda x: int(x.taxon.label)
    )
    assert {
        leaf_node.distance_from_root() for leaf_node in sorted_leaf_nodes
    } == {
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    }
    assert sorted(
        leaf_node.distance_from_root() for leaf_node in sorted_leaf_nodes
    ) == sorted(
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    )
    assert [
        leaf_node.distance_from_root() for leaf_node in sorted_leaf_nodes
    ] == [
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    ]

    for reconst_node_pair, extant_column_pair in zip(
        it.combinations(sorted_leaf_nodes, 2),
        it.combinations(extant_population, 2),
    ):
        reconst_mrca = impl.descend_unifurcations(
            pdm.mrca(*map(lambda x: x.taxon, reconst_node_pair))
        )

        (
            lower_mrca_bound,
            upper_mrca_bound,
        ) = hstrat.calc_rank_of_mrca_bounds_between(
            *extant_column_pair, prior="arbitrary"
        )

        assert (
            lower_mrca_bound
            <= reconst_mrca.distance_from_root()
            < upper_mrca_bound
        )


@pytest.mark.parametrize(
    "orig_tree",
    [
        pytest.param(
            impl.setup_dendropy_tree(f"{assets_path}/nk_ecoeaselection.csv"),
            marks=pytest.mark.heavy,
        ),
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
        impl.setup_dendropy_tree(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.recency_proportional_resolution_algo.Policy(3),
        hstrat.fixed_resolution_algo.Policy(5),
    ],
)
def test_col_specimen_consistency(orig_tree, retention_policy):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_instrument=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )

    reconst_df1 = build_tree_searchtable_python(
        extant_population,
    )
    reconst_df2 = build_tree_searchtable_python(
        [hstrat.col_to_specimen(col) for col in extant_population],
    )

    assert reconst_df1[["id", "ancestor_list"]].equals(
        reconst_df2[["id", "ancestor_list"]]
    )


@pytest.mark.parametrize("tree_size", [10, 30, 100, 300, 1000])
@pytest.mark.parametrize(
    "differentia_width", [1, pytest.param(2, marks=pytest.mark.heavy), 8, 64]
)
@pytest.mark.parametrize(
    "tree_seed",
    [*range(10)]
    + [pytest.param(i, marks=pytest.mark.heavy) for i in range(10, 100)],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.recency_proportional_resolution_algo.Policy(1),
        pytest.param(
            hstrat.recency_proportional_resolution_algo.Policy(2),
            marks=pytest.mark.heavy,
        ),
        pytest.param(
            hstrat.recency_proportional_resolution_algo.Policy(10),
            marks=pytest.mark.heavy,
        ),
        hstrat.fixed_resolution_algo.Policy(2),
    ],
)
@pytest.mark.parametrize(
    "exhaustive_check",
    [pytest.param(True, marks=pytest.mark.heavy), False],
)
def test_reconstructed_mrca_fuzz(
    tree_seed, tree_size, differentia_width, retention_policy, exhaustive_check
):

    seed_random(tree_seed)
    nx_tree = random_tree(n=tree_size, seed=tree_seed, create_using=nx.DiGraph)

    extant_population = hstrat.descend_template_phylogeny_networkx(
        nx_tree,
        seed_instrument=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
            stratum_differentia_bit_width=differentia_width,
        ),
    )

    reconst_df = build_tree_searchtable_python(
        extant_population,
        progress_wrap=tqdm,
    )
    assert "origin_time" in reconst_df

    assert pfl.alifestd_validate(reconst_df)
    assert pfl.alifestd_is_chronologically_ordered(reconst_df)
    reconst_tree = apc.alife_dataframe_to_dendropy_tree(
        reconst_df,
        setup_edge_lengths=True,
    )
    pdm = reconst_tree.phylogenetic_distance_matrix()

    assert len(list(reconst_tree.leaf_node_iter())) == len(extant_population)
    sorted_leaf_nodes = sorted(
        reconst_tree.leaf_node_iter(), key=lambda x: int(x.taxon.label)
    )
    assert {
        int(leaf_node.distance_from_root()) for leaf_node in sorted_leaf_nodes
    } == {
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    }
    assert sorted(
        int(leaf_node.distance_from_root()) for leaf_node in sorted_leaf_nodes
    ) == sorted(
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    )
    assert [
        int(leaf_node.distance_from_root()) for leaf_node in sorted_leaf_nodes
    ] == [
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    ]

    for reconst_node_pair, extant_column_pair, label_pair in (
        (
            (
                (sorted_leaf_nodes[first_idx], sorted_leaf_nodes[second_idx]),
                (extant_population[first_idx], extant_population[second_idx]),
                (first_idx, second_idx),
            )
            for first_idx, second_idx in generate_n(
                lambda: (
                    random.randrange(len(extant_population)),
                    random.randrange(len(extant_population)),
                ),
                10000,
            )
        )
        if not exhaustive_check
        else zip(
            it.combinations(sorted_leaf_nodes, 2),
            it.combinations(extant_population, 2),
            it.combinations(range(len(extant_population)), 2),
        )
    ):
        reconst_mrca = impl.descend_unifurcations(
            pdm.mrca(*map(lambda x: x.taxon, reconst_node_pair))
        )

        (
            lower_mrca_bound,
            upper_mrca_bound,
        ) = hstrat.calc_rank_of_mrca_bounds_between(
            *extant_column_pair,
            prior="arbitrary",
            confidence_level=0.99999,
            strict=False,
        )

        if not (
            lower_mrca_bound
            <= reconst_mrca.distance_from_root()
            < upper_mrca_bound
        ):
            print(reconst_mrca.taxon.label)
            print(reconst_mrca.distance_from_root())
            print()
            for i, (col, ln) in enumerate(
                zip(extant_population, sorted_leaf_nodes)
            ):
                print()
                print("label", i)
                print(ln.distance_from_root())
                print(id(col))
                print(hstrat.col_to_ascii(col))
            print(nx.forest_str(nx_tree))
            print(reconst_tree.as_ascii_plot(plot_metric="length"))

        assert (
            lower_mrca_bound
            <= reconst_mrca.distance_from_root()
            < upper_mrca_bound
        ), (
            label_pair[0],
            label_pair[1],
            id(extant_column_pair[0]),
            id(extant_column_pair[1]),
            lower_mrca_bound,
            reconst_mrca.distance_from_root(),
            upper_mrca_bound,
        )


@pytest.mark.parametrize(
    "orig_tree",
    [
        pytest.param(
            impl.setup_dendropy_tree(f"{assets_path}/nk_ecoeaselection.csv"),
            marks=pytest.mark.heavy,
        ),
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
        impl.setup_dendropy_tree(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.recency_proportional_resolution_algo.Policy(3),
        hstrat.fixed_resolution_algo.Policy(5),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [
        1,
        8,
        64,
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        hstrat.col_to_specimen,
    ],
)
def test_determinism(orig_tree, retention_policy, differentia_width, wrap):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_instrument=hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )

    first_reconst = build_tree_searchtable_python(
        extant_population,
    )
    for _rep in range(10):
        _ = _rep
        second_reconst = build_tree_searchtable_python(
            [wrap(col) for col in extant_population],
        )
        pd.testing.assert_frame_equal(first_reconst, second_reconst)


@pytest.mark.parametrize(
    "orig_tree",
    [
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.recency_proportional_resolution_algo.Policy(4),
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        hstrat.col_to_specimen,
    ],
)
def test_reconstructed_taxon_labels(orig_tree, retention_policy, wrap):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_instrument=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )
    taxon_labels = [str(id(x)) for x in extant_population]

    reconst_df = build_tree_searchtable_python(
        [*map(wrap, extant_population)],
        taxon_labels=taxon_labels,
    )
    assert "taxon_label" in reconst_df
    assert set(taxon_labels) <= set(reconst_df["taxon_label"])

    reconst_df = build_tree_searchtable_python(
        [*map(wrap, extant_population)],
    )
    assert "taxon_label" in reconst_df
    assert set(map(str, range(len(extant_population)))) <= set(
        reconst_df["taxon_label"]
    )



from deprecated.sphinx import deprecated
import numpy as np
import pandas as pd

from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_is_working_format_asexual import (
    alifestd_is_working_format_asexual,
)
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._jit import jit


@jit(nopython=True)
def _alifestd_calc_clade_trait_count_asexual_fast_path(
    ancestor_ids: np.ndarray,
    trait_mask: np.ndarray,
) -> np.ndarray:
    """Implementation detail for `alifestd_calc_clade_trait_count_asexual`."""

    # iterate over ancestor_ids from back to front
    for idx_r, ancestor_id in enumerate(ancestor_ids[::-1]):
        idx = len(ancestor_ids) - 1 - idx_r
        if ancestor_id == idx:
            continue  # handle genesis cases

        trait_mask[ancestor_id] += trait_mask[idx]

    return trait_mask


def _alifestd_calc_clade_trait_count_asexual_slow_path(
    phylogeny_df: pd.DataFrame,
) -> pd.DataFrame:
    """Implementation detail for `alifestd_calc_clade_trait_count_asexual`."""

    saved_id_order = None
    if not alifestd_is_topologically_sorted(phylogeny_df):
        saved_id_order = phylogeny_df["id"].to_numpy(copy=True)
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    phylogeny_df.index = phylogeny_df["id"]

    for idx in reversed(phylogeny_df.index):
        ancestor_id = phylogeny_df.at[idx, "ancestor_id"]
        if ancestor_id == idx:
            continue  # handle genesis cases

        phylogeny_df.at[
            ancestor_id, "alifestd_calc_trait_count_asexual"
        ] += phylogeny_df.at[idx, "alifestd_calc_trait_count_asexual"]

    if saved_id_order is not None:
        return phylogeny_df.loc[
            saved_id_order, "alifestd_calc_trait_count_asexual"
        ].to_numpy()
    else:
        return phylogeny_df["alifestd_calc_trait_count_asexual"].to_numpy()


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_calc_clade_trait_count_asexual instead.",
)
def alifestd_calc_clade_trait_count_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    *,
    trait_mask: np.ndarray,  # 1D boolean array
) -> np.ndarray:
    """Count how many nodes within each clade have a given trait.

    Clades are defined as a node and all descendant nodes.

    Returns a numpy array of the same length as the input DataFrame, with array
    elements as the number of nodes in the clade that have the trait. Returned
    array matches row order of the input DataFrame.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df["alifestd_calc_trait_count_asexual"] = np.asarray(
        trait_mask, dtype=int
    )
    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if alifestd_is_working_format_asexual(phylogeny_df):
        return _alifestd_calc_clade_trait_count_asexual_fast_path(
            pd.to_numeric(phylogeny_df["ancestor_id"]).to_numpy(),
            phylogeny_df["alifestd_calc_trait_count_asexual"].to_numpy(),
        )
    else:
        return _alifestd_calc_clade_trait_count_asexual_slow_path(phylogeny_df)



import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_coarsen_taxa_asexual,
    alifestd_make_empty,
)


@pytest.mark.parametrize("col", ["id", "ancestor_id", "ancestor_list"])
def test_coarsen_agg_override_error(col: str):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
            "dummy": [1, 1, 1],
        }
    )
    with pytest.raises(ValueError):
        alifestd_coarsen_taxa_asexual(
            phylogeny_df,
            agg={col: "sum"},
            by="dummy",
        )


@pytest.mark.parametrize("mutate", [True, False])
def test_empty_coarsen(mutate: bool):
    phylogeny_df = alifestd_make_empty()
    # add the column we intend to group by
    phylogeny_df["dummy"] = pd.Series(dtype=int)
    result = alifestd_coarsen_taxa_asexual(
        phylogeny_df,
        mutate=mutate,
        by="dummy",
    )
    assert len(result) == 0


@pytest.mark.parametrize("mutate", [True, False])
def test_singleton_coarsen(mutate: bool):
    # A single-node tree coarsens to itself.
    phylogeny_df = pd.DataFrame(
        {
            "id": [42],
            "ancestor_list": ["[None]"],
            "branch_length": [0],
            "edge_length": [0],
            "origin_time": [100],
            "destruction_time": [200],
        }
    )
    phylogeny_df["group"] = 7
    original = phylogeny_df.copy()

    result = alifestd_coarsen_taxa_asexual(
        phylogeny_df,
        mutate=mutate,
        by="group",
    )
    assert "branch_length" not in result
    assert "edge_length" not in result

    # Should get exactly one row, same id, and each override respected:
    assert result.shape == (1, len(result.columns))
    row = result.iloc[0]
    assert row["id"] == 42
    assert row["origin_time"] == 100
    assert row["destruction_time"] == 200

    if not mutate:
        pd.testing.assert_frame_equal(phylogeny_df, original)


@pytest.mark.parametrize("mutate", [True, False])
def test_chain_coarsen(mutate: bool):
    # Chain 0 -> 1 -> 2, all rows share dummy=0, so you get
    #   - one taxon at the root (id=0),
    #   - one taxon for the chain below (id=1, since we take 'first' for id).
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
            "branch_length": [0, 1, 2],
            "origin_time": [0, 10, 20],
        }
    )
    phylogeny_df["dummy"] = 0
    original = phylogeny_df.copy()

    result = alifestd_coarsen_taxa_asexual(
        phylogeny_df,
        mutate=mutate,
        by="dummy",
    )

    assert result["id"].tolist() == [0]
    assert result["ancestor_id"].tolist() == [0]
    assert "branch_length" not in result
    assert "edge_length" not in result

    if not mutate:
        pd.testing.assert_frame_equal(phylogeny_df, original)


@pytest.mark.parametrize("mutate", [True, False])
def test_star_coarsen(mutate: bool):
    # Star tree 0 -> {1,2}, all share dummy=0.
    # Depth=1 means each child winds up its own cluster,
    # so you get three rows back, with branch_length identical.
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[0]"],
            "branch_length": [0, 5, 5],
            "origin_time": [0, 50, 100],
        }
    )
    phylogeny_df["dummy"] = 0
    original = phylogeny_df.copy()

    result = alifestd_coarsen_taxa_asexual(
        phylogeny_df,
        mutate=mutate,
        by="dummy",
    )

    assert result["id"].tolist() == [0]
    assert result["ancestor_id"].tolist() == [0]

    if not mutate:
        pd.testing.assert_frame_equal(phylogeny_df, original)


@pytest.mark.parametrize("mutate", [True, False])
def test_group_by_sequence(mutate: bool):
    # Supplying `by` as a sequence should behave same as single key.
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
            "branch_length": [0, 1, 2],
            "origin_time": [0, 10, 20],
        }
    )
    phylogeny_df["g1"] = 0
    phylogeny_df["g2"] = 1
    original = phylogeny_df.copy()

    single = (
        alifestd_coarsen_taxa_asexual(phylogeny_df, mutate=mutate, by="g1")
        .sort_values("id")
        .reset_index(drop=True)
    )
    seq = (
        alifestd_coarsen_taxa_asexual(phylogeny_df, mutate=mutate, by=["g1"])
        .sort_values("id")
        .reset_index(drop=True)
    )

    pd.testing.assert_frame_equal(single, seq)

    if not mutate:
        pd.testing.assert_frame_equal(phylogeny_df, original)


@pytest.mark.parametrize("mutate", [True, False])
def test_chain_coarsen_with_overrides(mutate: bool):
    # Chain 0 -> 1 -> 2, all share group=0. Check that:
    #   - origin_time takes the first
    #   - destruction_time takes the last
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
            "branch_length": [0, 1, 2],
            "edge_length": [0, 2, 3],
            "origin_time": [0, 10, 20],
            "destruction_time": [5, 6, 7],
        }
    )
    phylogeny_df["group"] = [0, 1, 1]
    original = phylogeny_df.copy()

    result = alifestd_coarsen_taxa_asexual(
        phylogeny_df,
        mutate=mutate,
        by="group",
    )
    ids = sorted(result["id"])
    assert ids == [0, 1]
    assert "branch_length" not in result
    assert "edge_length" not in result

    agg = {row["id"]: row for _, row in result.iterrows()}
    # root cluster
    assert agg[0]["origin_time"] == 0
    assert agg[0]["destruction_time"] == 5
    assert agg[0]["ancestor_id"] == 0
    # lower cluster
    assert agg[1]["origin_time"] == 10
    assert agg[1]["destruction_time"] == 7
    assert agg[1]["ancestor_id"] == 0

    if not mutate:
        pd.testing.assert_frame_equal(phylogeny_df, original)


@pytest.mark.parametrize("mutate", [True, False])
def test_multiple_independent_chains(mutate: bool):
    # Two disjoint chains with group=0 or group=1.
    # Each chain collapses into two clusters.
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_list": [
                "[None]",
                "[0]",
                "[1]",
                "[None]",
                "[3]",
                "[4]",
            ],
            "branch_length": [0, 1, 2, 0, 3, 4],
            "origin_time": [0, 10, 20, 5, 15, 25],
        }
    )
    phylogeny_df["group"] = [0, 1, 1, 2, 3, 3]
    original = phylogeny_df.copy()

    result = alifestd_coarsen_taxa_asexual(
        phylogeny_df,
        mutate=mutate,
        by="group",
    )
    ids = sorted(result["id"])
    assert ids == [0, 1, 3, 4]

    assert "branch_length" not in result

    if not mutate:
        pd.testing.assert_frame_equal(phylogeny_df, original)


@pytest.mark.parametrize("mutate", [True, False])
def test_chain_two_group_keys(mutate: bool):
    # Chain 0 -> 1 -> 2, grouped by TWO keys:
    #   - g1 is constant
    #   - g2 flips from 0 -> 1 at node 1
    # So we should break into two clusters:
    #   - cluster at 0 alone
    #   - cluster of [1,2] together
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
            "branch_length": [0, 1, 2],
            "origin_time": [0, 10, 20],
        }
    )
    # g1 same throughout, g2 flips at node 1 and stays
    phylogeny_df["g1"] = [0, 0, 0]
    phylogeny_df["g2"] = [0, 1, 1]
    phylogeny_df["g3"] = ["A", "A", "A"]
    original = phylogeny_df.copy()

    result = alifestd_coarsen_taxa_asexual(
        phylogeny_df,
        mutate=mutate,
        by=["g1", "g2"],
    )
    # Expect two clusters: ids 0 and 1
    assert sorted(result["id"]) == [0, 1]
    assert "branch_length" not in result

    agg = {row["id"]: row for _, row in result.iterrows()}
    # root cluster
    assert agg[0]["ancestor_id"] == 0
    assert agg[0]["origin_time"] == 0
    # lower cluster
    assert agg[1]["ancestor_id"] == 0
    assert agg[1]["origin_time"] == 10

    if not mutate:
        pd.testing.assert_frame_equal(phylogeny_df, original)


@pytest.mark.parametrize("mutate", [True, False])
def test_branching_disconnected_same_group_with_parent_id(mutate: bool):
    # 10-node tree with single-parent pointers:
    #
    #       0
    #      / \
    #     1   2
    #    / \ / \
    #   3  4 5  6
    #  /      \  \
    # 7        8  9
    #
    # ancestor_id maps each node to exactly one parent (or None for root).
    # Groups:
    #   - 'root' on {0,1,2}
    #   - 'A' on two disconnected clades {3 -> 7} and {5 -> 8}
    #   - 'B' on two others {4} and {6 -> 9}
    #
    # Expect six clusters: one per contiguous segment per leaf path:
    #   - root-clade under 1        -> id=1
    #   - root-clade under 2        -> id=2
    #   - A-clade under 3 -> 7      -> id=7
    #   - A-clade under 5 ->        -> id=8
    #   - B-clade leaf 4            -> id=4
    #   - B-clade under 6 -> 9      -> id=9
    N = 10
    df = pd.DataFrame(
        {
            "id": list(range(N)),
            "ancestor_id": [0, 0, 0, 1, 1, 2, 2, 3, 5, 6],
            "branch_length": [1] * N,
            "edge_length": [2] * N,
            "origin_time": list(range(N)),
            "destruction_time": [i + 10 for i in range(N)],
        }
    )
    df["grp"] = [
        "root",  # 0
        "root",  # 1
        "root",  # 2
        "A",  # 3
        "root",  # 4
        "A",  # 5
        "B",  # 6
        "A",  # 7
        "A",  # 8
        "A",  # 9
    ]
    original = df.copy()

    result = alifestd_coarsen_taxa_asexual(
        df,
        mutate=mutate,
        by="grp",
    )

    # Expect exactly these terminal ids
    expected_ids = [0, 3, 5, 6, 9]
    assert sorted(result["id"].tolist()) == expected_ids

    # Check ancestor_ids
    anc = dict(zip(result["id"], result["ancestor_id"]))
    assert anc[0] == 0
    assert anc[3] == 0
    assert anc[5] == 0
    assert anc[6] == 0
    assert anc[9] == 6

    # Check origin_time = first in cluster, destruction_time = last
    ot = dict(zip(result["id"], result["origin_time"]))
    dt = dict(zip(result["id"], result["destruction_time"]))
    assert ot[0] == 0 and dt[0] == 14
    assert ot[3] == 3 and dt[3] == 17
    assert ot[5] == 5 and dt[5] == 18
    assert ot[6] == 6 and dt[6] == 16
    assert ot[9] == 9 and dt[9] == 19

    if not mutate:
        pd.testing.assert_frame_equal(df, original)


@pytest.mark.parametrize("mutate", [True, False])
def test_branching_disconnected_same_group_with_parent_id2(mutate: bool):
    # 10-node tree with single-parent pointers:
    #
    #       0
    #      / \
    #     1   2
    #    / \ / \
    #   3  4 5  6
    #  /      \  \
    # 7        8  99
    N = 10
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 99],
            "ancestor_id": [0, 0, 0, 1, 1, 2, 2, 3, 5, 6],
            "branch_length": [1] * N,
            "edge_length": [2] * N,
            "origin_time": list(range(N)),
            "destruction_time": [i + 10 for i in range(N)],
        }
    )
    df["grp"] = [
        "root",  # 0
        "root",  # 1
        "root",  # 2
        "A",  # 3
        "root",  # 4
        "A",  # 5
        "B",  # 6
        "A",  # 7
        "A",  # 8
        "A",  # 99
    ]
    df = df[::-1].copy()  # reverse (not topologically sorted)
    original = df.copy()

    result = alifestd_coarsen_taxa_asexual(
        df,
        mutate=mutate,
        by="grp",
    )

    # Expect exactly these terminal ids
    expected_ids = [0, 3, 5, 6, 99]
    assert sorted(result["id"].tolist()) == expected_ids

    # Check ancestor_ids
    anc = dict(zip(result["id"], result["ancestor_id"]))
    assert anc[0] == 0
    assert anc[3] == 0
    assert anc[5] == 0
    assert anc[6] == 0
    assert anc[99] == 6

    # Check origin_time = first in cluster, destruction_time = last
    ot = dict(zip(result["id"], result["origin_time"]))
    dt = dict(zip(result["id"], result["destruction_time"]))
    assert ot[0] == 0 and dt[0] == 14
    assert ot[3] == 3 and dt[3] == 17
    assert ot[5] == 5 and dt[5] == 18
    assert ot[6] == 6 and dt[6] == 16
    assert ot[99] == 9 and dt[99] == 19

    if not mutate:
        pd.testing.assert_frame_equal(df, original)



import argparse
from collections import Counter
import logging
import os

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import pandas as pd

from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._alifestd_try_add_ancestor_list_col import (
    alifestd_try_add_ancestor_list_col,
)
from ._begin_prod_logging import begin_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_topological_sort instead.",
)
def alifestd_topological_sort(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Sort rows so all organisms follow members of their `ancestor_list`.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    has_ancestor_list = "ancestor_list" in phylogeny_df
    if not has_ancestor_list:
        phylogeny_df = alifestd_try_add_ancestor_list_col(
            phylogeny_df, mutate=True
        )

    phylogeny_df.set_index("id", drop=False, inplace=True)

    unsorted_ids = set(phylogeny_df["id"])
    internal_ids_refcounts = Counter(
        id
        for ancestor_list_str in phylogeny_df["ancestor_list"]
        for id in alifestd_parse_ancestor_ids(ancestor_list_str)
    )
    internal_ids_set = set(internal_ids_refcounts.keys())

    reverse_sorted_ids = []
    while unsorted_ids:
        leaf_ids = unsorted_ids - internal_ids_set
        reverse_sorted_ids.extend(leaf_ids)
        unsorted_ids -= leaf_ids

        leaf_parent_counts = Counter(
            id
            for ancestor_list_str in phylogeny_df.loc[[*leaf_ids]][
                "ancestor_list"
            ]
            for id in alifestd_parse_ancestor_ids(ancestor_list_str)
        )
        internal_ids_refcounts.subtract(leaf_parent_counts)

        for leaf_parent_id in leaf_parent_counts.keys():
            refcount = internal_ids_refcounts[leaf_parent_id]
            assert refcount >= 0
            if refcount == 0:
                internal_ids_set.remove(leaf_parent_id)
                del internal_ids_refcounts[leaf_parent_id]

    assert set(reverse_sorted_ids) == set(phylogeny_df["id"])
    res = phylogeny_df.loc[reverse_sorted_ids[::-1], :].reset_index(drop=True)

    if not has_ancestor_list:
        res.drop(columns=["ancestor_list"], inplace=True)

    return res


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Sort rows so all organisms follow members of their `ancestor_list`.

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_topological_sort",
        dfcli_version=get_hstrat_version(),
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_topological_sort", logging.info
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                alifestd_topological_sort,
            ),
        )



import argparse
import functools
import logging
import os

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import numpy as np
import pandas as pd

from ._add_bool_arg import add_bool_arg
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_topological_sensitivity_warned import (
    alifestd_topological_sensitivity_warned,
)
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._alifestd_unfurl_lineage_asexual import alifestd_unfurl_lineage_asexual
from ._begin_prod_logging import begin_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration
from ._pairwise import pairwise


@alifestd_topological_sensitivity_warned(
    insert=False,
    delete=False,
    update=True,
)
@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_reroot_at_id_asexual instead.",
)
def alifestd_reroot_at_id_asexual(
    phylogeny_df: pd.DataFrame,
    new_root_id: int,
    mutate: bool = False,
) -> pd.DataFrame:
    """Reroot phylogeny, preserving topology.

    Reverses the descendant-to-ancestor relationships of all ancestors of the
    new root. Does not update branch_lengths or edge_lengths columns if
    present.

    Parameters
    ----------
    phylogeny_df : pandas.DataFrame
        The phylogeny as a dataframe in alife standard format.

        Must represent an asexual phylogeny.
    new_root_id : int
        The ID of the node to use as the new root of the phylogeny.
    mutate : bool, default False
        Are side effects on the input argument `phylogeny_df` allowed?

    Returns
    -------
    pandas.DataFrame
        The rerooted phylogeny in alife standard format.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)
    unfurled_lineage = alifestd_unfurl_lineage_asexual(
        phylogeny_df, new_root_id
    )

    # contiguous id implementation
    if alifestd_has_contiguous_ids(phylogeny_df):
        copy_to_slice = unfurled_lineage[1:]
        copy_from_slice = unfurled_lineage[:-1]
        phylogeny_df["ancestor_id"].to_numpy()[copy_to_slice] = phylogeny_df[
            "id"
        ].to_numpy()[copy_from_slice]

        phylogeny_df["ancestor_id"].to_numpy()[new_root_id] = phylogeny_df[
            "id"
        ].to_numpy()[new_root_id]

    # noncontiguous id implementation
    else:
        iloc_lookup = dict(
            zip(phylogeny_df["id"], np.arange(len(phylogeny_df)))
        )
        for ancestor_id, descendant_id in pairwise(reversed(unfurled_lineage)):
            iloc = iloc_lookup[ancestor_id]
            phylogeny_df["ancestor_id"].to_numpy()[iloc] = descendant_id

        new_root_iloc = iloc_lookup[new_root_id]
        phylogeny_df["ancestor_id"].to_numpy()[new_root_iloc] = phylogeny_df[
            "id"
        ].to_numpy()[new_root_iloc]

    # update ancestor list
    phylogeny_df["ancestor_list"] = alifestd_make_ancestor_list_col(
        phylogeny_df["id"],
        phylogeny_df["ancestor_id"],
    )
    return phylogeny_df


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Reroot phylogeny at specified node id, preserving topology.

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_reroot_at_id_asexual",
        dfcli_version=get_hstrat_version(),
    )
    parser.add_argument(
        "--new-root-id",
        type=int,
        required=True,
        help="id of the node to reroot at",
    )
    add_bool_arg(
        parser,
        "ignore-topological-sensitivity",
        default=False,
        help="suppress topological sensitivity warning (default: False)",
    )
    add_bool_arg(
        parser,
        "drop-topological-sensitivity",
        default=False,
        help="drop topology-sensitive columns from output (default: False)",
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_reroot_at_id_asexual", logging.info
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                functools.partial(
                    alifestd_reroot_at_id_asexual,
                    new_root_id=args.new_root_id,
                    ignore_topological_sensitivity=args.ignore_topological_sensitivity,
                    drop_topological_sensitivity=args.drop_topological_sensitivity,
                ),
            ),
        )



import logging

from deprecated.sphinx import deprecated
import numpy as np
import polars as pl

from ._alifestd_has_contiguous_ids_polars import (
    alifestd_has_contiguous_ids_polars,
)
from ._alifestd_mark_num_children_asexual import (
    _alifestd_mark_num_children_asexual_fast_path,
)


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_find_leaf_ids_polars instead.",
)
def alifestd_find_leaf_ids_polars(
    phylogeny_df: pl.DataFrame,
) -> np.ndarray:
    """What ids are ancestor to no other ids?

    Parameters
    ----------
    phylogeny_df : polars.DataFrame or polars.LazyFrame
        The phylogeny as a dataframe in alife standard format.

        Must have contiguous ids and represent an asexual phylogeny.

    Returns
    -------
    numpy.ndarray
        Array of leaf node ids.

    See Also
    --------
    alifestd_find_leaf_ids :
        Pandas-based implementation.
    """
    logging.info(
        "- alifestd_find_leaf_ids_polars: checking contiguous ids...",
    )
    if not alifestd_has_contiguous_ids_polars(phylogeny_df):
        raise NotImplementedError("non-contiguous ids not yet supported")

    logging.info(
        "- alifestd_find_leaf_ids_polars: extracting ancestor ids...",
    )
    ancestor_ids = (
        phylogeny_df.lazy()
        .select("ancestor_id")
        .collect()
        .to_series()
        .to_numpy()
    )

    logging.info(
        "- alifestd_find_leaf_ids_polars: tabulating child counts...",
    )
    child_counts = _alifestd_mark_num_children_asexual_fast_path(ancestor_ids)

    logging.info(
        "- alifestd_find_leaf_ids_polars: finding leaf ids...",
    )
    return np.flatnonzero(child_counts == 0)



import argparse
import functools
import logging
import os

import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli

from .. import hstrat
from .._auxiliary_lib import (
    add_bool_arg,
    begin_prod_logging,
    format_cli_description,
    get_hstrat_version,
    log_context_duration,
)
from ._surface_build_tree import surface_build_tree

raw_message = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

End-to-end interface to tree reconstruction for surface-based genome annotations.
Reads raw hex genome data from tabular data file(s) and writes postprocessed phylogeny data to output file in alife standard format.


Surface-only implementation --- not compatible with HereditaryStratigraphicColumn genome annotations.


Input Schema: Required Columns
==============================
'data_hex' : string
    Raw genome data, with serialized dstream buffer and counter.
    Represented as a hexadecimal string.

'dstream_algo' : string or categorical
    Name of downstream curation algorithm used (e.g., 'dstream.steady_algo').

'dstream_storage_bitoffset' : integer
    Position of dstream buffer field in 'data_hex'.

'dstream_storage_bitwidth' : integer
    Size of dstream buffer field in 'data_hex'.

'dstream_T_bitoffset' : integer
    Position of dstream counter field ("rank") in 'data_hex'.

'dstream_T_bitwidth' : integer
    Size of dstream counter field in 'data_hex'.

'dstream_S' : integer
    Capacity of dstream buffer, in number of data items (i.e., num differentia stored per annotation).


If required columns are not included in raw data, they must be created through CLI flags.
For example,
    --with-column 'pl.lit("dstream.tilted_algo").alias("dstream_algo").cast(pl.Categorical)'
    --with-column 'pl.lit("genome_string").alias("data_hex")'


To work with genome data in raw binary format (e.g., pl.Binary),
    --with-column 'pl.col("data_bin").bin.encode("hex").alias("data_hex")'


Additional user-provided columns will be forwarded to phylogeny output.
For these columns, output rows for tip nodes are assigned values from corresponding genome row in original data.
Internal tree nodes will take null values in user-provided columns.


Output Schema: Required Columns
===============================
'id' : integer
    Unique identifier for each taxon (RE alife standard format).
    Will be reassigned for contiguity and may differ from input.

'ancestor_id' : integer
    Unique identifier for ancestor taxon (RE alife standard format).

'hstrat_rank' : integer
    - Num generations elapsed for ancestral differentia.
    - Corresponds to `dstream_Tbar` - `dstream_S` for inner nodes.
    - Corresponds `dstream_T` - 1 - `dstream_S` for leaf nodes.


Output Schema: Optional Columns
===============================
'origin_time' : floating point or integer
    Estimated num generations elapsed from founding ancestor.
    Value depends on the trie postprocessor used.

Additional user-defined columns will be forwarded from the input DataFrame.
Note that the alife-standard `ancestor_list` column is not included in output schema.


See Also
========
-m hstrat.dataframe.surface_unpack_reconstrct :
-m hstrat.dataframe.surface_postprocess_trie :
    Low-level CLI interface for phylogeny reconstruction.
    See their docstrings/help messages for technical details on reconstruction
    and postprocessing operations.

-m hstrat._auxiliary_lib._alifestd_try_add_ancestor_list_col :
    Adds alife-standard `ancestor_list` column to phylogeny data.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(raw_message),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    _add_parser_base(
        parser=parser,
        dfcli_module="hstrat.dataframe.surface_build_tree",
        dfcli_version=get_hstrat_version(),
    )
    parser.add_argument(
        "--collapse-unif-freq",
        type=int,
        default=1,
        help="How often should dropped unifurcations be garbage collected?",
    )
    parser.add_argument(
        "--check-trie-invariant-freq",
        type=int,
        default=0,
        help=(
            "How often should trie invariant checks be run? "
            "Set to 0 to disable (default). "
            "Set to n > 0 to check every n slices."
        ),
    )
    add_bool_arg(
        parser,
        "check-trie-invariant-after-collapse-unif",
        default=False,
        help=(
            "Should trie invariant checks also be run after collapse "
            "unifurcations? Default is False (checks run before collapse only)."
        ),
    )
    parser.add_argument(
        "--exploded-slice-size",
        type=int,
        default=1_000_000,
        help="Number of rows to process at once. Low values reduce memory use.",
    )
    parser.add_argument(
        "--mp-pool-size",
        type=int,
        default=1,
        help=(
            "Number of worker processes for exploding slices in parallel. "
            "Default 1 (single producer process)."
        ),
    )
    parser.add_argument(
        "--delete-trunk",
        action=argparse.BooleanOptionalAction,
        default=True,
        help=(
            "Should trunk nodes with rank less than `dstream_S` be deleted? "
            "Trunk deletion accounts for 'dummy' strata added to fill hstrat "
            "surface for founding ancestor(s), by segregating subtrees with "
            "distinct founding strata into independent trees."
        ),
    )
    parser.add_argument(
        "--trie-postprocessor",
        type=str,
        default="hstrat.NopTriePostprocessor()",
        help=(
            "Functor to apply to finalize postprocessed phylogeny. "
            "Postprocessors are available to estimate taxon `origin_time`, "
            "apply topological corrections for systematic overestimation of "
            "relatedness, etc. "
            "Will be `eval`-ed as Python code before use. "
            "Must support Pandas dataframe input."
        ),
    )
    parser.add_argument(
        "--pa-source-type",
        type=str,
        default="memory_map",
        help=(
            "PyArrow type to use for exploded chunks "
            """(i.e., "memory_map" or "OSFile")."""
        ),
    )

    return parser


def _main(mp_context: str) -> None:
    parser = _create_parser()
    args, __ = parser.parse_known_args()

    logging.info(
        "instantiating trie postprocess functor: "
        f"`{args.trie_postprocessor}`",
    )
    trie_postprocessor = eval(args.trie_postprocessor, {"hstrat": hstrat})

    with log_context_duration(
        "hstrat.dataframe.surface_build_tree", logging.info
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=functools.partial(
                surface_build_tree,
                collapse_unif_freq=args.collapse_unif_freq,
                check_trie_invariant_freq=args.check_trie_invariant_freq,
                check_trie_invariant_after_collapse_unif=args.check_trie_invariant_after_collapse_unif,
                delete_trunk=args.delete_trunk,
                exploded_slice_size=args.exploded_slice_size,
                mp_context=mp_context,
                mp_pool_size=args.mp_pool_size,
                pa_source_type=args.pa_source_type,
                trie_postprocessor=trie_postprocessor,
            ),
        )


if __name__ == "__main__":
    begin_prod_logging()

    _main("spawn")



import typing

from deprecated.sphinx import deprecated
import numpy as np
import pandas as pd

from ._alifestd_is_working_format_asexual import (
    alifestd_is_working_format_asexual,
)
from ._alifestd_mark_node_depth_asexual import alifestd_mark_node_depth_asexual
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_calc_mrca_id_matrix_asexual instead.",
)
def alifestd_calc_mrca_id_matrix_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    *,
    progress_wrap: typing.Callable = lambda x: x,
) -> np.ndarray:
    """Calculate the Most Recent Common Ancestor (MRCA) taxon id for each pair
    of taxa.

    Taxa sharing no common ancestor will have MRCA id -1.

    Pass tqdm or equivalent as `progress_wrap` to display a progress bar.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if not alifestd_is_working_format_asexual(phylogeny_df, mutate=True):
        raise NotImplementedError(
            "current implementation requires phylogeny_df in working format",
        )

    phylogeny_df = alifestd_mark_node_depth_asexual(phylogeny_df, mutate=True)

    n = len(phylogeny_df)
    result = -np.ones((n, n), dtype=np.int64)
    if n == 0:
        return result

    max_depth = int(phylogeny_df["node_depth"].max())

    ancestor_ids = phylogeny_df["ancestor_id"].to_numpy()
    node_depths = phylogeny_df["node_depth"].to_numpy()
    cur_positions = phylogeny_df["id"].to_numpy().copy()
    assert np.all(cur_positions == np.arange(len(phylogeny_df)))

    for depth in progress_wrap(reversed(range(max_depth + 1))):
        depth_mask = node_depths[cur_positions] == depth

        ansatz = -np.ones_like(result)

        ansatz[:, depth_mask] = cur_positions[depth_mask]
        ansatz[depth_mask, :] = cur_positions[depth_mask, None]
        ansatz[~depth_mask, :] = -1
        ansatz[:, ~depth_mask] = -1

        ansatz[ansatz != ansatz.T] = -1

        result = np.maximum(result, ansatz)
        cur_positions[depth_mask] = ancestor_ids[cur_positions[depth_mask]]

    return result



import argparse
import contextlib
import functools
import logging
import os
import typing

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import numpy as np
import opytional as opyt
import pandas as pd
from tqdm import tqdm

from ._RngStateContext import RngStateContext
from ._add_bool_arg import add_bool_arg
from ._alifestd_calc_mrca_id_vector_asexual import (
    alifestd_calc_mrca_id_vector_asexual,
)
from ._alifestd_downsample_tips_lineage_asexual import (
    _alifestd_downsample_tips_lineage_select_target_id,
)
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_mark_leaves import alifestd_mark_leaves
from ._alifestd_prune_extinct_lineages_asexual import (
    alifestd_prune_extinct_lineages_asexual,
)
from ._alifestd_topological_sensitivity_warned import (
    alifestd_topological_sensitivity_warned,
)
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._begin_prod_logging import begin_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


def _alifestd_downsample_tips_lineage_stratified_impl(
    is_leaf: np.ndarray,
    criterion_values: np.ndarray,
    stratify_values: np.ndarray,
    mrca_vector: np.ndarray,
    n_tips: typing.Optional[int] = None,
    n_tips_per_stratum: int = 1,
) -> np.ndarray:
    """Shared numpy implementation for stratified lineage tip
    downsampling.

    Computes off-lineage deltas from a pre-computed MRCA vector and
    returns a boolean extant mask, selecting up to
    ``n_tips_per_stratum`` leaves with the smallest delta per
    stratified group.

    When `n_tips` is an integer, stratified values are coarsened by
    ranking and integer-dividing so that exactly
    ``n_tips // n_tips_per_stratum`` groups are formed.  When `n_tips`
    is ``None``, each distinct stratified value defines its own group.

    Parameters
    ----------
    is_leaf : numpy.ndarray
        Boolean array indicating which taxa are leaves.
    criterion_values : numpy.ndarray
        Values used to compute off-lineage delta (all taxa).
    stratify_values : numpy.ndarray
        Values used to stratify leaves into groups (all taxa).
    mrca_vector : numpy.ndarray
        Integer array of MRCA ids for each taxon with respect to the
        target leaf.  Taxa in a different tree should have ``-1``.
    n_tips : int, optional
        Desired number of retained tips.  If ``None``, every distinct
        stratified value forms its own group and
        ``n_tips_per_stratum`` tips are kept per group.
    n_tips_per_stratum : int, default 1
        Number of tips to retain per stratified group.

    Returns
    -------
    numpy.ndarray
        Boolean array of length ``len(is_leaf)`` marking retained taxa.
    """
    # Taxa with no common ancestor (different tree) get -1 from MRCA
    # calc; replace with a safe dummy id (0) so the lookup doesn't fail,
    # then exclude these taxa from selection below.
    no_mrca_mask = mrca_vector == -1
    safe_mrca = np.where(no_mrca_mask, 0, mrca_vector)

    logging.info(
        "_alifestd_downsample_tips_lineage_stratified_impl: "
        "calculating off lineage delta...",
    )
    off_lineage_delta = np.abs(
        criterion_values - criterion_values[safe_mrca],
    )

    # Select eligible leaves
    logging.info(
        "_alifestd_downsample_tips_lineage_stratified_impl: "
        "filtering leaf eligibility...",
    )
    is_eligible = is_leaf & ~no_mrca_mask
    eligible_ids = np.flatnonzero(is_eligible)
    eligible_deltas = off_lineage_delta[is_eligible]
    eligible_stratified = stratify_values[is_eligible]

    # Coarsen stratified values if n_tips is specified
    if n_tips is not None:
        logging.info(
            "_alifestd_downsample_tips_lineage_stratified_impl: "
            "coarsening stratified values...",
        )
        n_groups = n_tips // n_tips_per_stratum
        unique_sorted = np.unique(eligible_stratified)
        n_unique = len(unique_sorted)
        ranks = np.searchsorted(unique_sorted, eligible_stratified)
        eligible_stratified = ranks * n_groups // n_unique
        assert len(np.unique(eligible_stratified)) == min(n_groups, n_unique)

    # Per-stratum selection: for each group, keep up to
    # n_tips_per_stratum leaves with the smallest delta.
    logging.info(
        "_alifestd_downsample_tips_lineage_stratified_impl: "
        "selecting kept ids per stratum...",
    )
    # Vectorized: sort by (stratum, delta), then pick the first
    # n_tips_per_stratum occurrences of each stratum.
    sort_order = np.lexsort((eligible_deltas, eligible_stratified))
    sorted_strata = eligible_stratified[sort_order]
    sorted_ids = eligible_ids[sort_order]

    stratum_change = np.diff(sorted_strata, prepend=np.nan) != 0
    stratum_starts = np.flatnonzero(stratum_change)
    stratum_lengths = np.diff(
        np.append(stratum_starts, len(sorted_strata)),
    )
    start_for_each = np.repeat(stratum_starts, stratum_lengths)
    position_in_stratum = np.arange(len(sorted_strata)) - start_for_each
    keep_mask = position_in_stratum < n_tips_per_stratum
    kept_ids = sorted_ids[keep_mask]

    logging.info(
        "_alifestd_downsample_tips_lineage_stratified_impl: "
        "building extant mask...",
    )
    return np.bincount(kept_ids, minlength=len(is_leaf)).astype(bool)


@alifestd_topological_sensitivity_warned(
    insert=False,
    delete=True,
    update=False,
)
@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_downsample_tips_lineage_stratified_asexual instead.",
)
def alifestd_downsample_tips_lineage_stratified_asexual(
    phylogeny_df: pd.DataFrame,
    n_tips: typing.Optional[int] = None,
    mutate: bool = False,
    seed: typing.Optional[int] = None,
    *,
    criterion_delta: str = "origin_time",
    criterion_stratify: str = "origin_time",
    criterion_target: str = "origin_time",
    n_tips_per_stratum: int = 1,
    progress_wrap: typing.Callable = lambda x: x,
) -> pd.DataFrame:
    """Retain leaves per stratified group, chosen by proximity to the
    lineage of a target leaf.

    Selects a target leaf as the leaf with the largest `criterion_target`
    value (ties broken randomly). For each non-target leaf, the most
    recent common ancestor (MRCA) of that leaf and the target leaf is
    identified, and the "off-lineage delta" is computed as the absolute
    difference between that leaf's `criterion_delta` value and the
    MRCA's `criterion_delta` value.

    Leaves are grouped by their `criterion_stratify` value. When
    `n_tips` is an integer, stratified values are coarsened by ranking
    and integer-dividing to form exactly ``n_tips // n_tips_per_stratum``
    groups. When `n_tips` is ``None``, each distinct stratified value
    forms its own group (without ranking). Within each group, the
    ``n_tips_per_stratum`` leaves with the smallest off-lineage delta
    are retained.

    Only supports asexual phylogenies.

    Parameters
    ----------
    phylogeny_df : pandas.DataFrame
        The phylogeny as a dataframe in alife standard format.

        Must represent an asexual phylogeny.
    n_tips : int, optional
        Desired number of retained tips.  If ``None``, every distinct
        ``criterion_stratify`` value forms its own group.
    mutate : bool, default False
        Are side effects on the input argument `phylogeny_df` allowed?
    seed : int, optional
        Random seed for reproducible target-leaf selection when there are
        ties in `criterion_target`.
    criterion_delta : str, default "origin_time"
        Column name used to compute the off-lineage delta for each leaf.
        The delta is the absolute difference between a leaf's value and
        its MRCA's value in this column.
    criterion_stratify : str, default "origin_time"
        Column name used to stratify leaves into groups.
    criterion_target : str, default "origin_time"
        Column name used to select the target leaf. The leaf with the
        largest value in this column is chosen as the target. Note that
        ties are broken by random sample, allowing a seed to be
        provided.
    n_tips_per_stratum : int, default 1
        Number of tips to retain per stratified group.  Must evenly
        divide ``n_tips`` when ``n_tips`` is not ``None``.
    progress_wrap : Callable, optional
        Pass tqdm or equivalent to display a progress bar.

    Raises
    ------
    ValueError
        If `criterion_delta`, `criterion_stratify`, or
        `criterion_target` is not a column in `phylogeny_df`.
    ValueError
        If ``n_tips`` is not ``None`` and ``n_tips_per_stratum`` does
        not evenly divide ``n_tips``.

    Returns
    -------
    pandas.DataFrame
        The pruned phylogeny in alife standard format.
    """
    if n_tips is not None and n_tips % n_tips_per_stratum != 0:
        raise ValueError(
            f"n_tips_per_stratum={n_tips_per_stratum} does not evenly "
            f"divide n_tips={n_tips}",
        )

    for criterion in (
        criterion_delta,
        criterion_stratify,
        criterion_target,
    ):
        if criterion not in phylogeny_df.columns:
            raise ValueError(
                f"criterion column {criterion!r} not found in phylogeny_df",
            )

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if phylogeny_df.empty:
        return phylogeny_df

    logging.info(
        "- alifestd_downsample_tips_lineage_stratified_asexual: "
        "adding ancestor_id col...",
    )
    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df)
    if "ancestor_id" not in phylogeny_df.columns:
        raise ValueError(
            "alifestd_downsample_tips_lineage_stratified_asexual only "
            "supports asexual phylogenies.",
        )

    logging.info(
        "- alifestd_downsample_tips_lineage_stratified_asexual: "
        "marking leaves...",
    )
    if "is_leaf" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_leaves(phylogeny_df)

    logging.info(
        "- alifestd_downsample_tips_lineage_stratified_asexual: "
        "checking contiguous ids...",
    )
    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        # non-contiguous id branch not tested because
        # alifestd_calc_mrca_id_vector_asexual doesn't support it
        raise NotImplementedError(
            "non-contiguous ids not yet supported",
        )

    is_leaf = phylogeny_df["is_leaf"].to_numpy()
    target_values = phylogeny_df[criterion_target].to_numpy()
    criterion_values = phylogeny_df[criterion_delta].to_numpy()
    stratify_values = phylogeny_df[criterion_stratify].to_numpy()

    with opyt.apply_if_or_else(seed, RngStateContext, contextlib.nullcontext):
        target_id = _alifestd_downsample_tips_lineage_select_target_id(
            is_leaf, target_values
        )

    logging.info(
        "- alifestd_downsample_tips_lineage_stratified_asexual: "
        "computing mrca vector...",
    )
    mrca_vector = alifestd_calc_mrca_id_vector_asexual(
        phylogeny_df, target_id=target_id, progress_wrap=progress_wrap
    )

    logging.info(
        "- alifestd_downsample_tips_lineage_stratified_asexual: "
        "computing lineage stratified downsample...",
    )
    is_extant = _alifestd_downsample_tips_lineage_stratified_impl(
        is_leaf=is_leaf,
        criterion_values=criterion_values,
        stratify_values=stratify_values,
        mrca_vector=mrca_vector,
        n_tips=n_tips,
        n_tips_per_stratum=n_tips_per_stratum,
    )

    logging.info(
        "- alifestd_downsample_tips_lineage_stratified_asexual: pruning...",
    )
    phylogeny_df["extant"] = is_extant
    return alifestd_prune_extinct_lineages_asexual(
        phylogeny_df, mutate=True
    ).drop(columns=["extant"])


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Retain one leaf per stratified group, chosen by proximity to the
lineage of a target leaf.

The target leaf is chosen as the leaf with the largest
`--criterion-target` value. For each leaf, the off-lineage delta is
the absolute difference between that leaf's `--criterion-delta` value
and the MRCA's `--criterion-delta` value (where the MRCA is of that
leaf and the target).
Leaves are grouped by their `--criterion-stratify` value. When `-n`
is given, stratified values are coarsened into `-n` groups by ranking
and integer division. Within each group, the leaf with the smallest
delta is retained.

Data is assumed to be in alife standard format.
Only supports asexual phylogenies.

Additional Notes
================
- Requires 'ancestor_id' column to be present in input DataFrame.
Otherwise, no action is taken.

- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_downsample_tips_lineage_stratified_asexual",
        dfcli_version=get_hstrat_version(),
    )
    parser.add_argument(
        "-n",
        default=None,
        type=int,
        help="Number of stratified groups (default: one per distinct value).",
    )
    parser.add_argument(
        "--criterion-delta",
        default="origin_time",
        type=str,
        help="Column used to compute off-lineage delta (default: origin_time).",
    )
    parser.add_argument(
        "--criterion-stratify",
        default="origin_time",
        type=str,
        help="Column used to stratify leaves (default: origin_time).",
    )
    parser.add_argument(
        "--criterion-target",
        default="origin_time",
        type=str,
        help="Column used to select the target leaf (default: origin_time).",
    )
    parser.add_argument(
        "--n-tips-per-stratum",
        default=1,
        type=int,
        help="Number of tips per stratum (default: 1).",
    )
    parser.add_argument(
        "--seed",
        default=None,
        dest="seed",
        help="Integer seed for deterministic target-leaf selection.",
        type=int,
    )
    add_bool_arg(
        parser,
        "ignore-topological-sensitivity",
        default=False,
        help="suppress topological sensitivity warning (default: False)",
    )
    add_bool_arg(
        parser,
        "drop-topological-sensitivity",
        default=False,
        help="drop topology-sensitive columns from output (default: False)",
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_downsample_tips_lineage_stratified_asexual",
        logging.info,
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                functools.partial(
                    alifestd_downsample_tips_lineage_stratified_asexual,
                    n_tips=args.n,
                    seed=args.seed,
                    criterion_delta=args.criterion_delta,
                    criterion_stratify=args.criterion_stratify,
                    criterion_target=args.criterion_target,
                    n_tips_per_stratum=args.n_tips_per_stratum,
                    progress_wrap=tqdm,
                    ignore_topological_sensitivity=args.ignore_topological_sensitivity,
                    drop_topological_sensitivity=args.drop_topological_sensitivity,
                ),
            ),
            overridden_arguments="ignore",  # seed is overridden
        )



import argparse
import logging
import os

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import pandas as pd

from ._alifestd_find_chronological_inconsistency import (
    alifestd_find_chronological_inconsistency,
)
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._alifestd_topological_sort import alifestd_topological_sort
from ._begin_prod_logging import begin_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_coerce_chronological_consistency instead.",
)
def alifestd_coerce_chronological_consistency(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """For any taxa with origin time preceding its parent's, set origin time
    to parent's origin time.

    If an inconsistency is detected, the corrected phylogeny will be returned
    sorted in topological order.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if (
        not alifestd_is_topologically_sorted(phylogeny_df)
        and alifestd_find_chronological_inconsistency(phylogeny_df) is not None
    ):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    phylogeny_df = phylogeny_df.set_index("id", drop=False)

    # adapted from https://stackoverflow.com/a/31569794
    origin_time_loc = phylogeny_df.columns.get_loc("origin_time")
    for pos, (_idx, row) in enumerate(phylogeny_df.iterrows()):
        for ancestor_id in alifestd_parse_ancestor_ids(row["ancestor_list"]):
            ancestor_loc = phylogeny_df.index.get_loc(ancestor_id)
            ancestor_time = phylogeny_df.iat[ancestor_loc, origin_time_loc]
            offspring_time = phylogeny_df.iat[pos, origin_time_loc]
            phylogeny_df.iat[pos, origin_time_loc] = max(
                offspring_time, ancestor_time
            )

    return phylogeny_df


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

For any taxa with origin time preceding its parent's, set origin time to parent's origin time.

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_coerce_chronological_consistency",
        dfcli_version=get_hstrat_version(),
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_coerce_chronological_consistency",
        logging.info,
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                alifestd_coerce_chronological_consistency,
            ),
        )



import argparse
import logging
import os

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import pandas as pd

from ._alifestd_find_leaf_ids import alifestd_find_leaf_ids
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._begin_prod_logging import begin_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_mark_leaves instead.",
)
def alifestd_mark_leaves(
    phylogeny_df: pd.DataFrame, mutate: bool = False
) -> pd.DataFrame:
    """What rows are ancestor to no other row?

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    leaf_ids = alifestd_find_leaf_ids(phylogeny_df)
    phylogeny_df["is_leaf"] = False
    phylogeny_df.loc[leaf_ids, "is_leaf"] = True

    return phylogeny_df


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Mark rows that are ancestor to no other row with an `is_leaf` column.

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.

See Also
========
hstrat._auxiliary_lib._alifestd_mark_leaves_polars :
    Entrypoint for high-performance Polars-based implementation.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_mark_leaves",
        dfcli_version=get_hstrat_version(),
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_mark_leaves", logging.info
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                alifestd_mark_leaves,
            ),
        )



import os

import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_aggregate_phylogenies,
    alifestd_count_leaf_nodes,
    alifestd_find_leaf_ids,
    alifestd_prune_extinct_lineages_asexual,
    alifestd_try_add_ancestor_id_col,
    alifestd_validate,
)
from hstrat._auxiliary_lib._alifestd_downsample_tips_canopy_asexual import (
    alifestd_downsample_tips_canopy_asexual,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize("num_tips", [1, 5, 10, 100000000])
@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_downsample_tips_canopy_asexual(
    phylogeny_df, num_tips, mutate
):
    original_df = phylogeny_df.copy()

    result_df = alifestd_downsample_tips_canopy_asexual(
        phylogeny_df, num_tips, mutate, criterion="id"
    )

    assert len(result_df) <= len(original_df)
    assert "extant" not in result_df.columns

    if not mutate:
        pd.testing.assert_frame_equal(phylogeny_df, original_df)

    assert all(result_df["id"].isin(original_df["id"]))
    assert alifestd_count_leaf_nodes(result_df) == min(
        alifestd_count_leaf_nodes(original_df), num_tips
    )


@pytest.mark.parametrize("num_tips", [0, 1])
def test_alifestd_downsample_tips_canopy_asexual_with_zero_tips(num_tips):
    phylogeny_df = pd.DataFrame({"id": [], "parent_id": [], "ancestor_id": []})

    result_df = alifestd_downsample_tips_canopy_asexual(
        phylogeny_df, num_tips, criterion="id"
    )

    assert result_df.empty


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize("num_tips", [1, 5, 10])
def test_downsample_canopy_vs_manual(phylogeny_df, num_tips):
    """Verify canopy prune matches manually marking top tips as extant and
    then pruning extinct lineages."""
    original_df = phylogeny_df.copy()
    canopy_df = alifestd_downsample_tips_canopy_asexual(
        phylogeny_df, num_tips, mutate=False, criterion="id"
    )

    # manually replicate: find top tips, mark extant, prune
    tips = alifestd_find_leaf_ids(original_df)
    tips_sorted = np.sort(tips)
    kept = tips_sorted[-num_tips:]

    original_df = alifestd_try_add_ancestor_id_col(original_df)
    original_df["extant"] = original_df["id"].isin(kept)
    manual_df = alifestd_prune_extinct_lineages_asexual(
        original_df, mutate=True
    ).drop(columns=["extant"])

    pd.testing.assert_frame_equal(canopy_df, manual_df)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_alifestd_downsample_tips_canopy_asexual_retains_highest_ids(
    phylogeny_df,
):
    """Verify that the retained tips are the ones with the highest ids."""
    num_tips = 5
    result_df = alifestd_downsample_tips_canopy_asexual(
        phylogeny_df, num_tips, criterion="id"
    )

    original_tips = alifestd_find_leaf_ids(phylogeny_df)
    expected_kept = set(sorted(original_tips)[-num_tips:])

    result_tips = set(alifestd_find_leaf_ids(result_df))
    assert result_tips == expected_kept


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_alifestd_downsample_tips_canopy_asexual_validates(phylogeny_df):
    num_tips = 5
    result_df = alifestd_downsample_tips_canopy_asexual(
        phylogeny_df, num_tips, criterion="id"
    )
    assert alifestd_validate(result_df)


def test_alifestd_downsample_tips_canopy_asexual_simple():
    """Test a simple hand-crafted tree.

    Tree structure:
        0 (root)
        +-- 1
        |   +-- 3 (leaf)
        |   +-- 4 (leaf)
        +-- 2 (leaf)

    With num_tips=2, keep leaves 3 and 4 (highest ids), result is 0, 1, 3, 4.
    With num_tips=1, keep leaf 4 (highest id), result is 0, 1, 4.
    """
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[none]", "[0]", "[0]", "[1]", "[1]"],
        }
    )

    result2 = alifestd_downsample_tips_canopy_asexual(df, 2, criterion="id")
    assert set(result2["id"]) == {0, 1, 3, 4}

    result1 = alifestd_downsample_tips_canopy_asexual(df, 1, criterion="id")
    assert set(result1["id"]) == {0, 1, 4}


def test_alifestd_downsample_tips_canopy_asexual_all_tips():
    """Requesting more tips than exist should return the full phylogeny."""
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[none]", "[0]", "[0]", "[1]", "[1]"],
        }
    )

    result = alifestd_downsample_tips_canopy_asexual(
        df, 100000, criterion="id"
    )
    assert len(result) == 5


def test_alifestd_downsample_tips_canopy_asexual_tied_criterion():
    """When all leaves share the same criterion value, exactly num_tips
    should still be retained (ties broken arbitrarily)."""
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[none]", "[0]", "[0]", "[1]", "[1]"],
            "time": [0, 0, 0, 0, 0],
        }
    )
    # leaves are 2, 3, 4 — all have time=0
    for num_tips in (1, 2, 3):
        result = alifestd_downsample_tips_canopy_asexual(
            df, num_tips, criterion="time"
        )
        assert alifestd_count_leaf_nodes(result) == num_tips


def test_alifestd_downsample_tips_canopy_asexual_missing_criterion():
    """Verify ValueError when criterion column is missing."""
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[none]", "[0]", "[0]"],
        }
    )

    with pytest.raises(ValueError, match="criterion column"):
        alifestd_downsample_tips_canopy_asexual(df, 1, criterion="nonexistent")



import typing

from astropy.utils.decorators import deprecated_renamed_argument
import numpy as np
import pandas as pd
from phyloframe import legacy as pfl

from hstrat._auxiliary_lib import HereditaryStratigraphicInstrument

from ..._auxiliary_lib import (
    cast_int_lossless,
    give_len,
    unfurl_lineage_with_contiguous_ids,
)
from ..._auxiliary_lib._alifestd_assign_contiguous_ids import (
    _reassign_ids_asexual,
)
from ._descend_template_phylogeny import descend_template_phylogeny


@deprecated_renamed_argument("seed_column", "seed_instrument", since="1.20.0")
def descend_template_phylogeny_alifestd(
    phylogeny_df: pd.DataFrame,
    seed_instrument: HereditaryStratigraphicInstrument,
    extant_ids: typing.Optional[typing.Iterable[int]] = None,
    progress_wrap: typing.Callable = lambda x: x,
) -> typing.List[HereditaryStratigraphicInstrument]:
    """Generate a population of hereditary stratigraphic instruments that could
    have resulted from the template phylogeny.

    Parameters
    ----------
    phylogeny_df : pandas.DataFrame
        Phylogeny record as an alife-standard DataFrame.
    seed_instrument : HereditaryStratigraphicInstrument
        Hereditary stratigraphic instrument to seed at root node of phylogeny.

        Returned hereditary stratigraphic column population will be generated
        as if repeatedly calling `CloneDescendant()` on `seed_instrument`. As
        such, specifies configuration (i.e., differentia bit width and stratum
        retention policy) for returned columns. May already have strata
        deposited, which will be incorporated into generated extant population.
    extant_ids : optional list of int
        Which organisms should hereditary stratigraphic instruments be created
        for?

        Designates content and order of returned list of hereditary
        stratigraphic instrument, with id values corresponding to "id"
        instrument in `phylogeny_df`.

        If None, hereditary stratigraphic instruments will be created for all
        phylogenetic leaves (organisms without offspring) in order of
        appearance in `phylogeny_df`.
    progress_wrap : Callable, optional
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.

    Returns
    -------
    list of HereditaryStratigraphicInstrument
        Population of hereditary stratigraphic instruments for extant lineage
        members (i.e., phylogeny leaf nodes).

        Instruments ordered in order of appearance of corresponding extant
        organism id.
    """

    assert not pfl.alifestd_has_multiple_roots(phylogeny_df)

    # must take leaf_ids before possible topological sort to preserve order
    if extant_ids is None:
        extant_ids = pfl.alifestd_find_leaf_ids(phylogeny_df)

    working_df = pfl.alifestd_to_working_format(phylogeny_df)
    assert "ancestor_id" in working_df

    if not pfl.alifestd_has_contiguous_ids(phylogeny_df):
        if not pfl.alifestd_is_topologically_sorted(phylogeny_df):
            phylogeny_df = pfl.alifestd_topological_sort(phylogeny_df)

        extant_ids = _reassign_ids_asexual(
            phylogeny_df["id"].to_numpy(), np.fromiter(extant_ids, int)
        )

    ancestor_id_lookup = working_df["ancestor_id"].iloc

    def lookup_ancestor_id(id_: int) -> int:
        return ancestor_id_lookup[id_]

    def get_stem_length(id_: int) -> int:
        ancestor_id = lookup_ancestor_id(id_)
        if ancestor_id == id_:
            return 0
        elif "origin_time" in working_df:
            res = (
                working_df["origin_time"].iloc[id_]
                - working_df["origin_time"].iloc[ancestor_id]
            )
            assert res >= 0
            return cast_int_lossless(
                res, action="warn", context="origin time difference"
            )
        elif "generation" in working_df:
            res = (
                working_df["generation"].iloc[id_]
                - working_df["generation"].iloc[ancestor_id]
            )
            assert res >= 0
            return cast_int_lossless(
                res, action="warn", context="origin time difference"
            )
        else:
            return 1

    return descend_template_phylogeny(
        give_len(
            (
                unfurl_lineage_with_contiguous_ids(
                    working_df["ancestor_id"].to_numpy(),
                    leaf_id,
                )
                for leaf_id in extant_ids
            ),
            len(extant_ids),
        ),
        working_df["id"].to_numpy(),  # descending_tree_iterable
        lookup_ancestor_id,
        get_stem_length,
        seed_instrument=seed_instrument,
        demark=lambda x: x,
        progress_wrap=progress_wrap,
    )



import os
import typing

import pandas as pd
import polars as pl
import pytest

from hstrat._auxiliary_lib import (
    alifestd_is_topologically_sorted,
    alifestd_to_working_format,
)
from hstrat._auxiliary_lib._alifestd_is_topologically_sorted_polars import (
    alifestd_is_topologically_sorted_polars,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        alifestd_to_working_format(
            pd.read_csv(
                f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
            )
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_tournamentselection.csv")
        ),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_is_topologically_sorted_polars_true(
    phylogeny_df: pd.DataFrame, apply: typing.Callable
):
    """Topologically sorted + contiguous ids should return True."""
    df = apply(pl.from_pandas(phylogeny_df))
    assert alifestd_is_topologically_sorted_polars(df)


@pytest.mark.parametrize(
    "ids,ancestor_ids",
    [
        ([0, 1, 2], [0, 2, 0]),  # node 1 has ancestor 2 > 1
        ([0, 1, 2, 3], [0, 0, 3, 0]),  # node 2 has ancestor 3 > 2
        ([0, 1, 2, 3, 4], [0, 0, 0, 4, 1]),  # node 3 has ancestor 4 > 3
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_is_topologically_sorted_polars_false(
    ids: typing.List, ancestor_ids: typing.List, apply: typing.Callable
):
    """Data with sorted ids but invalid topological order should be False."""
    df = apply(
        pl.DataFrame({"id": ids, "ancestor_id": ancestor_ids}),
    )
    assert not alifestd_is_topologically_sorted_polars(df)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        alifestd_to_working_format(
            pd.read_csv(
                f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
            )
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")
        ),
        alifestd_to_working_format(
            pd.read_csv(f"{assets_path}/nk_tournamentselection.csv")
        ),
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_is_topologically_sorted_polars_matches_pandas(
    phylogeny_df: pd.DataFrame, apply: typing.Callable
):
    """Verify polars result matches pandas result for sorted input."""
    result_pd = alifestd_is_topologically_sorted(phylogeny_df)

    df = apply(pl.from_pandas(phylogeny_df))
    result_pl = alifestd_is_topologically_sorted_polars(df)

    assert result_pd == result_pl


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_is_topologically_sorted_polars_simple_sorted(
    apply: typing.Callable,
):
    """Simple chain 0 -> 1 -> 2 is topologically sorted."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 0, 1],
            }
        ),
    )
    assert alifestd_is_topologically_sorted_polars(df)


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_is_topologically_sorted_polars_simple_unsorted(
    apply: typing.Callable,
):
    """Node 1 has ancestor_id 2 which is > 1, so not topologically sorted."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 2, 0],
            }
        ),
    )
    assert not alifestd_is_topologically_sorted_polars(df)


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_is_topologically_sorted_polars_simple_tree(
    apply: typing.Callable,
):
    """Test a simple tree.

    Tree structure:
        0 (root)
        +-- 1
        |   +-- 3
        |   +-- 4
        +-- 2
    """
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2, 3, 4],
                "ancestor_id": [0, 0, 0, 1, 1],
            }
        ),
    )
    assert alifestd_is_topologically_sorted_polars(df)


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_is_topologically_sorted_polars_single_node(
    apply: typing.Callable,
):
    """A single root node is topologically sorted."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0],
                "ancestor_id": [0],
            }
        ),
    )
    assert alifestd_is_topologically_sorted_polars(df)


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_is_topologically_sorted_polars_two_roots(
    apply: typing.Callable,
):
    """Two independent roots, sorted."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2, 3],
                "ancestor_id": [0, 1, 0, 1],
            }
        ),
    )
    assert alifestd_is_topologically_sorted_polars(df)


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_is_topologically_sorted_polars_all_roots(
    apply: typing.Callable,
):
    """All self-referencing roots are topologically sorted."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_id": [0, 1, 2],
            }
        ),
    )
    assert alifestd_is_topologically_sorted_polars(df)


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_is_topologically_sorted_polars_empty(apply: typing.Callable):
    """Empty dataframe is topologically sorted."""
    df = apply(
        pl.DataFrame(
            {"id": [], "ancestor_id": []},
            schema={"id": pl.Int64, "ancestor_id": pl.Int64},
        ),
    )
    assert alifestd_is_topologically_sorted_polars(df)


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_is_topologically_sorted_polars_no_ancestor_id(
    apply: typing.Callable,
):
    """Verify NotImplementedError for missing ancestor_id."""
    df = apply(
        pl.DataFrame(
            {
                "id": [0, 1, 2],
                "ancestor_list": ["[none]", "[0]", "[1]"],
            }
        ),
    )
    with pytest.raises(NotImplementedError):
        alifestd_is_topologically_sorted_polars(df)


@pytest.mark.parametrize(
    "apply",
    [
        pytest.param(lambda x: x, id="DataFrame"),
        pytest.param(lambda x: x.lazy(), id="LazyFrame"),
    ],
)
def test_alifestd_is_topologically_sorted_polars_unsorted_ids(
    apply: typing.Callable,
):
    """Verify NotImplementedError for unsorted id values."""
    df = apply(
        pl.DataFrame(
            {
                "id": [2, 0, 1],
                "ancestor_id": [0, 0, 0],
            }
        ),
    )
    with pytest.raises(NotImplementedError):
        alifestd_is_topologically_sorted_polars(df)



import pandas as pd
import pandas.testing as pdt
import pytest

from hstrat._auxiliary_lib import (
    alifestd_add_inner_leaves,
    alifestd_make_empty,
)


def test_empty_df():
    df = alifestd_make_empty()
    result = alifestd_add_inner_leaves(df)
    assert result.empty


def test_single_node_df():
    df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_id": [0],
            "origin_time": [0],
            "is_root": [True],
        },
    )
    result = alifestd_add_inner_leaves(df)

    expected = pd.DataFrame(
        {
            "id": [0],
            "ancestor_id": [0],
            "origin_time": [0],
            "is_leaf": [True],
            "is_root": [True],
        },
    )
    pdt.assert_frame_equal(result, expected, check_like=True)


@pytest.mark.parametrize("mutate", [False, True])
def test_internal_node_add_leaves(mutate: bool):
    # Simple chain: 0 -> 1 -> 2
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_id": [0, 0, 1],
            "is_root": [True, False, False],
            "origin_time_delta": [1, 1, 1],
        },
    )
    original = df.copy()
    result = alifestd_add_inner_leaves(df, mutate=mutate)
    assert mutate or df.equals(original)

    expected = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_id": [0, 0, 1, 0, 1],
            "is_leaf": [False, False, True, True, True],
            "is_root": [True, False, False, False, False],
            "origin_time_delta": [1, 1, 1, 0, 0],
        },
    )
    pdt.assert_frame_equal(result, expected, check_like=True)


@pytest.mark.parametrize("mutate", [False, True])
def test_ancestor_id_only(mutate: bool):
    df_id = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_id": [0, 0, 0],
            "is_leaf": [False, True, True],
            "is_root": [True, False, False],
        }
    )

    original = df_id.copy()
    result = alifestd_add_inner_leaves(df_id, mutate=mutate)
    assert mutate or original.equals(df_id)

    expected_id = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 0, 0],
            "is_leaf": [False, True, True, True],
            "is_root": [True, False, False, False],
        }
    )

    pdt.assert_frame_equal(result, expected_id)


@pytest.mark.parametrize("mutate", [False, True])
def test_add_inner_leaves_ancestor_list(mutate: bool):
    df_list = pd.DataFrame(
        {
            "id": [0, 1],
            "ancestor_id": [0, 0],
            "ancestor_list": ["[none]", "[0]"],
            "is_leaf": [False, True],
            "is_root": [True, False],
        }
    )

    original = df_list.copy()
    result = alifestd_add_inner_leaves(df_list, mutate=mutate)
    assert mutate or original.equals(df_list)

    expected_list = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_id": [0, 0, 0],
            "ancestor_list": ["[none]", "[0]", "[0]"],
            "is_leaf": [False, True, True],
            "is_root": [True, False, False],
        }
    )

    pdt.assert_frame_equal(result.reset_index(drop=True), expected_list)


@pytest.mark.parametrize("mutate", [False, True])
def test_add_inner_leaves_ancestor_list_only1(mutate: bool):
    df_list = pd.DataFrame(
        {
            "id": [0, 1],
            "ancestor_list": ["[none]", "[none]"],
        }
    )

    original = df_list.copy()
    result = alifestd_add_inner_leaves(df_list, mutate=False)
    assert mutate or original.equals(df_list)

    expected_list = pd.DataFrame(
        {
            "id": [0, 1],
            "ancestor_list": ["[none]", "[none]"],
            "is_leaf": [True, True],
        }
    )

    pdt.assert_frame_equal(result, expected_list)


@pytest.mark.parametrize("mutate", [False, True])
def test_add_inner_leaves_ancestor_list_only2(mutate: bool):
    df_list = pd.DataFrame(
        {
            "id": [0, 1],
            "ancestor_list": ["[none]", "[0]"],
        }
    )

    original = df_list.copy()
    result = alifestd_add_inner_leaves(df_list, mutate=False)
    assert mutate or original.equals(df_list)

    expected_list = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[none]", "[0]", "[0]"],
            "is_leaf": [False, True, True],
        }
    )

    pdt.assert_frame_equal(result, expected_list)



import logging

from ._configure_prod_logging import configure_prod_logging
from ._get_hstrat_version import get_hstrat_version


def begin_prod_logging() -> None:
    configure_prod_logging()
    logging.info(f"hstrat v{get_hstrat_version()}")



import argparse
import functools
import gc
import logging
import os

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import polars as pl

from ._add_bool_arg import add_bool_arg
from ._alifestd_has_contiguous_ids_polars import (
    alifestd_has_contiguous_ids_polars,
)
from ._alifestd_is_topologically_sorted_polars import (
    alifestd_is_topologically_sorted_polars,
)
from ._alifestd_prune_extinct_lineages_asexual import (
    _create_has_extant_descendant_contiguous_sorted,
)
from ._alifestd_topological_sensitivity_warned_polars import (
    alifestd_topological_sensitivity_warned_polars,
)
from ._begin_prod_logging import begin_prod_logging
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration
from ._log_memory_usage import log_memory_usage


@alifestd_topological_sensitivity_warned_polars(
    insert=False,
    delete=True,
    update=False,
)
@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_prune_extinct_lineages_polars instead.",
)
def alifestd_prune_extinct_lineages_polars(
    phylogeny_df: pl.DataFrame,
) -> pl.DataFrame:
    """Drop taxa without extant descendants.

    An "extant" column, if provided, is used to determine extant taxa.
    Otherwise, taxa with inf or nan "destruction_time" are considered extant.

    Parameters
    ----------
    phylogeny_df : polars.DataFrame
        The phylogeny as a dataframe in alife standard format.

        Must represent an asexual phylogeny.

    Raises
    ------
    NotImplementedError
        If `phylogeny_df` has no "ancestor_id" column.
    NotImplementedError
        If `phylogeny_df` has non-contiguous ids.
    NotImplementedError
        If `phylogeny_df` is not topologically sorted.
    ValueError
        If `phylogeny_df` has neither "extant" nor "destruction_time" columns.

    Returns
    -------
    polars.DataFrame
        The pruned phylogeny in alife standard format.

    See Also
    --------
    alifestd_prune_extinct_lineages_asexual :
        Pandas-based implementation.
    """
    logging.info(
        "- alifestd_prune_extinct_lineages_polars: collecting schema...",
    )
    schema_names = phylogeny_df.lazy().collect_schema().names()
    if "ancestor_id" not in schema_names:
        raise NotImplementedError("ancestor_id column required")
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_prune_extinct_lineages_polars: checking empty...",
    )
    if phylogeny_df.lazy().limit(1).collect().is_empty():
        return phylogeny_df
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_prune_extinct_lineages_polars: "
        "checking contiguous ids...",
    )
    if not alifestd_has_contiguous_ids_polars(phylogeny_df):
        raise NotImplementedError("non-contiguous ids not yet supported")
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_prune_extinct_lineages_polars: "
        "checking topological sort...",
    )
    if not alifestd_is_topologically_sorted_polars(phylogeny_df):
        raise NotImplementedError(
            "polars topological sort not yet implemented",
        )
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_prune_extinct_lineages_polars: "
        "determining extant mask...",
    )
    if "extant" in schema_names:
        extant_mask = phylogeny_df.lazy().select("extant")
    elif "destruction_time" in schema_names:
        extant_mask = phylogeny_df.lazy().select(
            ~pl.col("destruction_time").is_finite()
        )
    else:
        raise ValueError('Need "extant" or "destruction_time" column.')
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_prune_extinct_lineages_polars: "
        "collecting extant mask...",
    )
    extant_mask = extant_mask.cast(pl.Boolean).collect().to_series().to_numpy()
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_prune_extinct_lineages_polars: "
        "collecting ancestor_ids...",
    )
    ancestor_ids = (
        phylogeny_df.lazy()
        .select("ancestor_id")
        .cast(pl.UInt64)
        .collect()
        .to_series()
        .to_numpy()
    )
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_prune_extinct_lineages_polars: "
        "calculating has_extant_descendant...",
    )
    has_extant_descendant = _create_has_extant_descendant_contiguous_sorted(
        ancestor_ids.copy(),  # must copy to remove read-only flag...
        extant_mask.copy(),  # ... for numba compatibility
    )
    del ancestor_ids, extant_mask
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_prune_extinct_lineages_polars: marking...",
    )
    phylogeny_df = phylogeny_df.with_columns(
        alifestd_has_extant_descendant=has_extant_descendant,
    )
    del has_extant_descendant
    gc.collect()
    log_memory_usage(logging.info)

    logging.info(
        "- alifestd_prune_extinct_lineages_polars: filtering...",
    )
    return phylogeny_df.filter(
        pl.col("alifestd_has_extant_descendant"),
    ).drop("alifestd_has_extant_descendant")


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Drop taxa without extant descendants.

An "extant" column, if provided, is used to determine extant taxa.
Otherwise, taxa with inf or nan "destruction_time" are considered extant.

Data is assumed to be in alife standard format.
Only supports asexual phylogenies.

Additional Notes
================
- Requires 'ancestor_id' column to be present in input DataFrame.
Otherwise, no action is taken.

- Requires 'extant' or 'destruction_time' column.

- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.

See Also
========
hstrat._auxiliary_lib._alifestd_prune_extinct_lineages_asexual :
    CLI entrypoint for Pandas-based implementation.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_prune_extinct_lineages_polars",
        dfcli_version=get_hstrat_version(),
    )
    add_bool_arg(
        parser,
        "ignore-topological-sensitivity",
        default=False,
        help="suppress topological sensitivity warning (default: False)",
    )
    add_bool_arg(
        parser,
        "drop-topological-sensitivity",
        default=False,
        help="drop topology-sensitive columns from output (default: False)",
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()

    try:
        with log_context_duration(
            "hstrat._auxiliary_lib._alifestd_prune_extinct_lineages_polars",
            logging.info,
        ):
            _run_dataframe_cli(
                base_parser=parser,
                output_dataframe_op=functools.partial(
                    alifestd_prune_extinct_lineages_polars,
                    ignore_topological_sensitivity=args.ignore_topological_sensitivity,
                    drop_topological_sensitivity=args.drop_topological_sensitivity,
                ),
            )
    except NotImplementedError as e:
        logging.error(
            "- polars op not yet implemented, use pandas op CLI instead",
        )
        raise e



from deprecated.sphinx import deprecated
import numpy as np
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_calc_clade_lookback_n_asexual instead.",
)
def alifestd_calc_clade_lookback_n_asexual(
    phylogeny_df: pd.DataFrame,
    lookback_n: int,
    mutate: bool = False,
) -> np.ndarray:
    """Find ancestor ids of nodes that are `lookback_n` nodes away in the
    phylogeny.

    The root node will be returned if the lookback distance exceeds available
    nodes.

    Returns a numpy array of the same length as the input DataFrame, with array
    elements as the number of nodes in the clade that have the trait. Returned
    array matches row order of the input DataFrame.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    lookback_result = phylogeny_df["id"].values.copy()
    for __ in range(lookback_n):
        lookback_result = phylogeny_df.loc[
            lookback_result, "ancestor_id"
        ].values.copy()

    return lookback_result



from __future__ import annotations

from copy import deepcopy
import random
import typing

from downstream import dsurf
import opytional as opyt

from ._HereditaryStratum import HereditaryStratum


class HereditaryStratigraphicSurface:
    """Wrapper around `downstream.dsurf.Surface` that supports the same
    interface as `HereditaryStratigraphicColumn`.

    Assumes that surface is initialized filled with differentiae, meaning that
    dstream T values are offset from rank/number of items deposited by surface
    size S. This also means that preinitialized-differentia have negative rank.
    """

    __slots__ = (
        "_differentia_bit_width",
        "_surface",
    )

    # how many bits wide of differentia should the deposited strata be
    # constructed with?
    _differentia_bit_width: int

    # stores the actual downstream surface
    _surface: dsurf.Surface[int]

    def __init__(
        self: "HereditaryStratigraphicSurface",
        dstream_surface: dsurf.Surface[int],
        *,
        predeposit_strata: typing.Optional[int] = None,
        stratum_differentia_bit_width: int = 64,
    ) -> None:
        """A wrapper around the downstream Surface object to match the
        `hstrat.HereditaryStratigraphicColumn` interface.

        Initially depoists `S + predeposit_strata` strata, where `S` is the
        surface size. If surface is already initialized, `predeposit_strata` is
        ignored and a ValueError is raised if `predeposit_strata` is not None.

        Parameters
        ----------
        dstream_surface: downstream.dsurf.Surface
            The surface to use to store annotations.
        predeposit_strata: int, default 1
            The number of strata to deposit on the surface during
            initialization. Default 1 matches behavior of
            HereditaryStratigraphicColumn.
        stratum_differentia_bit_width : int, default 64
            The bit width for generated differentia.
        """
        self._surface = dstream_surface
        self._differentia_bit_width = stratum_differentia_bit_width
        if dstream_surface.T == 0:
            predeposit_strata = opyt.or_value(predeposit_strata, 1)
            self.DepositStrata(self._surface.S + predeposit_strata)
            assert self.GetNextRank() == predeposit_strata
        elif dstream_surface.T < self._surface.S:
            raise ValueError("Partially-initialized Surface provided.")
        elif predeposit_strata is not None:
            raise ValueError(
                "Predeposit strata provided to already-initialized "
                "Surface. Predeposit strata should only be provided to "
                "uninitialized surfaces."
            )

    @property
    def S(self):
        """Number of buffer sites available to store differentiae."""
        return self._surface.S

    def __eq__(
        self: "HereditaryStratigraphicSurface",
        other: typing.Any,
    ) -> bool:
        """Compare two HereditaryStratigraphicSurface objects by value."""
        if not isinstance(other, HereditaryStratigraphicSurface):
            return False
        return (
            self._surface == other._surface
            and self._differentia_bit_width == other._differentia_bit_width
        )

    def _CreateStratum(
        self: "HereditaryStratigraphicSurface",
        deposition_rank: typing.Optional[int] = None,  # noqa: ARG001
        # ^ no-op for column interface compat
        *,
        differentia: typing.Optional[int] = None,
    ) -> int:
        """Create a hereditary stratum with stored configuration attributes."""
        return opyt.or_value(
            differentia,
            random.randrange(2**self._differentia_bit_width),
        )

    def DepositStratum(
        self: "HereditaryStratigraphicSurface",
        differentia: typing.Optional[int] = None,
    ) -> None:
        """Elapse a generation by depositing a differentia value.

        Differentia may be provided (useful for testing) or, by default,
        randomly generated.
        """
        if (
            differentia is not None
            and int(differentia).bit_length() > self._differentia_bit_width
        ):
            raise ValueError(
                f"Differentia {differentia} is too large for "
                f"{self._differentia_bit_width} bit width.",
            )
        self._surface.ingest_one(
            opyt.or_else(differentia, self._CreateStratum),
        )

    def DepositStrata(
        self: "HereditaryStratigraphicSurface",
        num_stratum_depositions: int,
    ) -> None:
        """Elapse n generations.

        Parameters
        ----------
        num_stratum_depositions: int
            How many generations to elapse?
        """
        self._surface.ingest_many(
            num_stratum_depositions, lambda _: self._CreateStratum()
        )

    def IterRetainedRanks(
        self: "HereditaryStratigraphicSurface",
    ) -> typing.Iterator[int]:
        """Iterate over deposition ranks of strata stored in the surface.

        The store may be altered during iteration without iterator
        invalidation, although subsequent updates will not be reflected in the
        iterator.
        """
        return (r for r, __ in self.IterRankDifferentiaZip())

    def IterRetainedStrata(
        self: "HereditaryStratigraphicSurface",
    ) -> typing.Iterator[HereditaryStratum]:
        """Iterate over strata stored in the surface.

        Strata are yielded from most ancient to most recent.
        """
        return (
            HereditaryStratum(
                annotation=None,
                deposition_rank=r,
                differentia_bit_width=self._differentia_bit_width,
                differentia=d,
            )
            for r, d in self.IterRankDifferentiaZip()
        )

    def IterRetainedDifferentia(
        self: "HereditaryStratigraphicSurface",
    ) -> typing.Iterator[int]:
        """Iterate over differentia of strata stored in the surface.

        Differentia yielded from most ancient to most recent.
        """
        return (d for __, d in self.IterRankDifferentiaZip())

    def IterRankDifferentiaZip(
        self: "HereditaryStratigraphicSurface",
        copyable: bool = False,  # noqa: ARG001 ... no-op for interface compat
    ) -> typing.Iterator[typing.Tuple[int, int]]:
        """Iterate over ranks of retained strata and their differentia.

        If `copyable`, return an iterator that can be copied to produce a new
        fully-independent iterator at the same position.

        Equivalent to `zip(surf.IterRetainedRanks(),
        surf.IterRetainedDifferentia())`.
        """

        return iter(
            sorted(
                (T - self._surface.S, v)
                for T, v in self._surface.lookup_zip_items(include_empty=False)
            ),
        )

    def HasAnyAnnotations(
        self: "HereditaryStratigraphicSurface",
    ) -> bool:
        """Do any retained strata have annotations?"""
        return False  # annotation feature not supported

    def GetNextRank(self: "HereditaryStratigraphicSurface") -> int:
        """Get the rank of the next stratum to be deposited.

        This is the rank of the next stratum to be deposited, not the rank of
        the most recent stratum.
        """
        return self._surface.T - self._surface.S

    def GetNumStrataRetained(self: "HereditaryStratigraphicSurface") -> int:
        """How many strata are currently stored within the surface?

        May be fewer than the number of strata deposited if strata have been
        discarded as part of the configured stratum retention policy.
        """
        return min(self._surface.T, self._surface.S)

    def GetNumStrataDeposited(self: "HereditaryStratigraphicSurface") -> int:
        """How many strata have been depostited on the surface?

        Note that a first `S` strata may be deposited on the surface during
        initialization.
        """
        return self._surface.T

    def GetStratumAtColumnIndex(
        self: "HereditaryStratigraphicSurface",
        index: int,
    ) -> HereditaryStratum:
        """Get the stratum positioned at index i among retained strata.

        Index order is from most ancient (index 0) to most recent.
        """
        return [*self.IterRetainedStrata()][index]

    def GetStratumAtRank(
        self: "HereditaryStratigraphicSurface",
        rank: int,
    ) -> typing.Optional[HereditaryStratum]:
        """Get the stratum deposited at generation g.

        Returns None if stratum with rank g is not retained.
        """
        for s in self.IterRetainedStrata():
            if s.GetDepositionRank() == rank:
                return s
        return None

    def GetRankAtStorageIndex(
        self: "HereditaryStratigraphicSurface",
        index: int,
    ) -> typing.Optional[int]:
        """Map column position to generation of deposition.

        What is the deposition rank of the stratum positioned at index i
        among retained strata? Index order is from most ancient (index 0) to
        most recent.
        """
        r = [*self._surface.lookup(include_empty=True)][index]
        return r if r is None else r - self._surface.S

    def GetRankAtColumnIndex(
        self: "HereditaryStratigraphicSurface",
        index: int,
    ) -> int:
        """Map column position to generation of deposition.

        What is the deposition rank of the stratum positioned at index i
        among retained strata? Index order is from most ancient (index 0) to
        most recent.
        """
        return [*self.IterRetainedRanks()][index]

    def GetStorageIndexOfRank(
        self: "HereditaryStratigraphicSurface",
        rank: int,
    ) -> typing.Optional[int]:
        """Map generation of deposition to position within surface storage.

        What is the surface position within retained strata of the stratum
        deposited at rank r? Returns None if no stratum with rank r is present
        within the store.
        """
        for i, r in enumerate(self._surface.lookup(include_empty=True)):
            if r is not None and r == rank + self._surface.S:
                return i
        return None

    def GetColumnIndexOfRank(
        self: "HereditaryStratigraphicSurface",
        rank: int,
    ) -> typing.Optional[int]:
        """Map generation of deposition to column position.

        What is the index position within retained strata of the stratum
        deposited at rank r? Returns None if no stratum with rank r is present
        within the store.
        """
        for i, r in enumerate(self.IterRetainedRanks()):
            if r == rank:
                return i
        return None

    def GetNumDiscardedStrata(
        self: "HereditaryStratigraphicSurface",
    ) -> int:
        """How many deposited strata have been discarded?

        Determined by number of generations elapsed and the configured column
        retention policy. Does not include `S` initialization strata.
        """
        return max(
            self.GetNumStrataDeposited()
            - self.GetNumStrataRetained()
            - self.S,
            0,
        )

    def GetStratumDifferentiaBitWidth(
        self: "HereditaryStratigraphicSurface",
    ) -> int:
        """How many bits wide are the differentia of strata?"""
        return self._differentia_bit_width

    def HasDiscardedStrata(
        self: "HereditaryStratigraphicSurface",
    ) -> bool:
        """Have any deposited strata been discarded?

        Does not include `S` initialization strata."""
        return self.GetNumDiscardedStrata() > 0

    def Clone(
        self: "HereditaryStratigraphicSurface",
    ) -> "HereditaryStratigraphicSurface":
        """Create an independent copy of the surface.

        Contains identical data but may be freely altered without affecting
        data within this surface.
        """
        return deepcopy(self)

    def CloneDescendant(
        self: "HereditaryStratigraphicSurface",
    ) -> "HereditaryStratigraphicSurface":
        """Return a cloned surface that has had an additional stratum deposited.

        Does not alter self.
        """
        new = self.Clone()
        new.DepositStratum()
        return new

    def CloneNthDescendant(
        self: "HereditaryStratigraphicSurface",
        num_stratum_depositions: int,
    ) -> "HereditaryStratigraphicSurface":
        """Return a cloned surface that has had n additional strata deposited.

        Does not alter self.

        Parameters
        ----------
        num_stratum_depositions: int
            How many generations should clone surface be descended?
        """
        new = self.Clone()
        new.DepositStrata(num_stratum_depositions)
        return new



import argparse
import logging
import os

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import polars as pl

from ._alifestd_mark_num_children_polars import (
    alifestd_mark_num_children_polars,
)
from ._begin_prod_logging import begin_prod_logging
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_mark_leaves_polars instead.",
)
def alifestd_mark_leaves_polars(
    phylogeny_df: pl.DataFrame,
) -> pl.DataFrame:
    """Add column `is_leaf` marking rows that are ancestor to no other row.

    Parameters
    ----------
    phylogeny_df : polars.DataFrame
        The phylogeny as a dataframe in alife standard format.

        Must represent an asexual phylogeny.

    Returns
    -------
    polars.DataFrame
        The phylogeny with an added `is_leaf` boolean column.

    See Also
    --------
    alifestd_mark_leaves :
        Pandas-based implementation.
    """
    if phylogeny_df.lazy().limit(1).collect().is_empty():
        return phylogeny_df.with_columns(
            is_leaf=pl.lit(False),  # sets Boolean dtype, but no values
        )

    if "num_children" not in phylogeny_df.lazy().collect_schema().names():
        phylogeny_df = alifestd_mark_num_children_polars(phylogeny_df)

    return phylogeny_df.with_columns(
        is_leaf=pl.col("num_children") == 0,
    )


_raw_description = f"""{os.path.basename(__file__)} | (hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Mark rows that are ancestor to no other row with an `is_leaf` column.

Data is assumed to be in alife standard format.

Additional Notes
================
- Requires 'ancestor_id' column to be present in input DataFrame.
Otherwise, no action is taken.

- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.

See Also
========
hstrat._auxiliary_lib._alifestd_mark_leaves :
    CLI entrypoint for Pandas-based implementation.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_mark_leaves_polars",
        dfcli_version=get_hstrat_version(),
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()

    try:
        with log_context_duration(
            "hstrat._auxiliary_lib._alifestd_mark_leaves_polars",
            logging.info,
        ):
            _run_dataframe_cli(
                base_parser=parser,
                output_dataframe_op=alifestd_mark_leaves_polars,
            )
    except NotImplementedError as e:
        logging.error(
            "- polars op not yet implemented, use pandas op CLI instead",
        )
        raise e



from deprecated.sphinx import deprecated
import numpy as np
import pandas as pd


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_has_contiguous_ids instead.",
)
def alifestd_has_contiguous_ids(phylogeny_df: pd.DataFrame) -> bool:
    """Do organisms ids' correspond to their row number?

    Input dataframe is not mutated by this operation.
    """
    return (
        phylogeny_df["id"].to_numpy() == np.arange(len(phylogeny_df))
    ).all()



import argparse
import logging
import os

from deprecated.sphinx import deprecated
import joinem
from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import numpy as np
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_strictly_bifurcating_asexual import (
    alifestd_is_strictly_bifurcating_asexual,
)
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_mark_left_child_asexual import alifestd_mark_left_child_asexual
from ._alifestd_mark_num_leaves_asexual import alifestd_mark_num_leaves_asexual
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._begin_prod_logging import begin_prod_logging
from ._delegate_polars_implementation import delegate_polars_implementation
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._jit import jit
from ._log_context_duration import log_context_duration


@jit(nopython=True)
def alifestd_mark_colless_index_asexual_fast_path(
    ancestor_ids: np.ndarray,
    num_leaves: np.ndarray,
    left_child_ids: np.ndarray,
) -> np.ndarray:
    """Implementation detail for
    `alifestd_mark_colless_index_asexual`.
    """
    n = len(ancestor_ids)

    # Compute local colless for each internal node
    # For bifurcating: right_leaves = num_leaves[node] - num_leaves[left]
    local_colless = np.zeros(n, dtype=np.int64)
    for tup in enumerate(left_child_ids):
        # workaround: unpack inside loop body to avoid numba
        # generator-in-zip KeyError bug with `for a, b in generator`
        idx, left_child = tup
        if left_child != idx:  # Has children (internal node)
            left_leaves = num_leaves[left_child]
            right_leaves = num_leaves[idx] - left_leaves
            local_colless[idx] = abs(left_leaves - right_leaves)

    # Reverse pass: accumulate subtree colless bottom-up
    colless_index = np.zeros(n, dtype=np.int64)
    for idx in range(n - 1, -1, -1):  # reversed enumerate
        ancestor_id = ancestor_ids[idx]
        colless_index[idx] += local_colless[idx]
        if ancestor_id != idx:  # Not a root
            colless_index[ancestor_id] += colless_index[idx]

    return colless_index


def alifestd_mark_colless_index_asexual_slow_path(
    phylogeny_df: pd.DataFrame,
) -> np.ndarray:
    """Implementation detail for
    `alifestd_mark_colless_index_asexual`.
    """
    phylogeny_df.index = phylogeny_df["id"]

    # Compute local colless using left_child approach
    local_colless = {}
    for node_id in phylogeny_df.index:
        left_child = phylogeny_df.at[node_id, "left_child_id"]
        if left_child != node_id:  # Has children (internal node)
            left_leaves = phylogeny_df.at[left_child, "num_leaves"]
            right_leaves = phylogeny_df.at[node_id, "num_leaves"] - left_leaves
            local_colless[node_id] = abs(left_leaves - right_leaves)
        else:
            local_colless[node_id] = 0

    # Accumulate subtree Colless (bottom-up via reversed iteration)
    for node_id in reversed(phylogeny_df.index):
        ancestor_id = phylogeny_df.at[node_id, "ancestor_id"]
        if ancestor_id != node_id:  # Not a root
            local_colless[ancestor_id] += local_colless[node_id]

    return phylogeny_df["id"].map(local_colless).values


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_mark_colless_index_asexual instead.",
)
def alifestd_mark_colless_index_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add column `colless_index` with Colless imbalance index for
    each subtree.

    Computes the classic Colless index for strictly bifurcating trees.
    For each internal node with exactly two children, the local
    contribution is |L - R| where L and R are leaf counts in left and
    right subtrees. The value at each node represents the total Colless
    index for the subtree rooted at that node.

    Raises ValueError if the tree is not strictly bifurcating. For
    trees with polytomies, use
    `alifestd_mark_colless_like_index_mdm_asexual` for the Colless-like
    index instead.

    Leaf nodes will have Colless index 0 (no imbalance in subtree of
    size 1). The root node contains the Colless index for the entire
    tree.

    A topological sort will be applied if `phylogeny_df` is not
    topologically sorted. Dataframe reindexing (e.g., df.index) may
    be applied.

    Input dataframe is not mutated by this operation unless `mutate`
    set True. If mutate set True, operation does not occur in place;
    still use return value to get transformed phylogeny dataframe.

    Parameters
    ----------
    phylogeny_df : pd.DataFrame
        Alife standard DataFrame containing the phylogenetic
        relationships.

    mutate : bool, optional
        If True, modify the input DataFrame in place. Default is
        False.

    Returns
    -------
    pd.DataFrame
        Phylogeny DataFrame with an additional column "colless_index"
        containing the Colless imbalance index for the subtree rooted
        at each node.

    Raises
    ------
    ValueError
        If phylogeny_df is not strictly bifurcating.

    See Also
    --------
    alifestd_mark_colless_index_corrected_asexual :
        Normalized Colless index (corrected for tree size).
    alifestd_mark_colless_like_index_mdm_asexual :
        Colless-like index (MDM) that supports polytomies.
    alifestd_mark_colless_like_index_var_asexual :
        Colless-like index (variance) that supports polytomies.
    alifestd_mark_colless_like_index_sd_asexual :
        Colless-like index (std dev) that supports polytomies.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if phylogeny_df.empty:
        phylogeny_df["colless_index"] = pd.Series(dtype=int)
        return phylogeny_df

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)
    if "ancestor_id" not in phylogeny_df.columns:
        raise ValueError(
            "alifestd_mark_colless_index_asexual only supports "
            "asexual phylogenies.",
        )

    if not alifestd_is_strictly_bifurcating_asexual(phylogeny_df):
        raise ValueError(
            "phylogeny_df must be strictly bifurcating; "
            "consider using "
            "alifestd_mark_colless_like_index_mdm_asexual "
            "for the Colless-like index for trees with polytomies",
        )

    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    if "num_leaves" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_num_leaves_asexual(
            phylogeny_df, mutate=True
        )

    if "left_child_id" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_left_child_asexual(
            phylogeny_df, mutate=True
        )

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df[
            "colless_index"
        ] = alifestd_mark_colless_index_asexual_fast_path(
            phylogeny_df["ancestor_id"].to_numpy(),
            phylogeny_df["num_leaves"].to_numpy(),
            phylogeny_df["left_child_id"].to_numpy(),
        )
    else:
        phylogeny_df[
            "colless_index"
        ] = alifestd_mark_colless_index_asexual_slow_path(
            phylogeny_df,
        )

    return phylogeny_df


_raw_description = f"""\
{os.path.basename(__file__)} | \
(hstrat v{get_hstrat_version()}/joinem v{joinem.__version__})

Add column `colless_index` with Colless imbalance index for each subtree.

Requires strictly bifurcating trees.

Data is assumed to be in alife standard format.

Additional Notes
================
- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        add_help=False,
        allow_abbrev=False,
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser = _add_parser_base(
        parser=parser,
        dfcli_module=(
            "hstrat._auxiliary_lib._alifestd_mark_colless_index_asexual"
        ),
        dfcli_version=get_hstrat_version(),
    )
    return parser


if __name__ == "__main__":
    begin_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()
    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_mark_colless_index_asexual",
        logging.info,
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=delegate_polars_implementation()(
                alifestd_mark_colless_index_asexual,
            ),
        )



import functools
import typing

from deprecated.sphinx import deprecated
import polars as pl

from ._alifestd_drop_topological_sensitivity_polars import (
    alifestd_drop_topological_sensitivity_polars,
)
from ._alifestd_warn_topological_sensitivity_polars import (
    alifestd_warn_topological_sensitivity_polars,
)


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_topological_sensitivity_warned_polars instead.",
)
def alifestd_topological_sensitivity_warned_polars(
    *, insert: bool, delete: bool, update: bool
) -> typing.Callable:
    """Decorator that emits a topological sensitivity warning before the
    wrapped function executes.

    The first positional argument of the decorated function must be the
    phylogeny dataframe (polars).

    The decorated function gains two additional keyword arguments:

    - ``ignore_topological_sensitivity`` (bool, default False):
      If True, suppress the topological sensitivity warning.
    - ``drop_topological_sensitivity`` (bool, default False):
      If True, drop topology-sensitive columns from the result and
      suppress the warning.

    Parameters
    ----------
    insert : bool
        Whether the operation inserts new nodes.
    delete : bool
        Whether the operation deletes nodes.
    update : bool
        Whether the operation updates ancestor relationships.

    Returns
    -------
    typing.Callable
        A decorator that wraps a function with topological sensitivity
        warning logic.

    See Also
    --------
    alifestd_topological_sensitivity_warned :
        Pandas-based implementation.
    alifestd_warn_topological_sensitivity_polars :
        Underlying warning function.
    """

    def decorator(func: typing.Callable) -> typing.Callable:
        @functools.wraps(func)
        def wrapper(
            phylogeny_df: typing.Union[pl.DataFrame, pl.LazyFrame],
            *args: typing.Any,
            ignore_topological_sensitivity: bool = False,
            drop_topological_sensitivity: bool = False,
            **kwargs: typing.Any,
        ) -> typing.Union[pl.DataFrame, pl.LazyFrame]:
            if (
                not ignore_topological_sensitivity
                and not drop_topological_sensitivity
            ):
                alifestd_warn_topological_sensitivity_polars(
                    phylogeny_df,
                    func.__name__,
                    insert=insert,
                    delete=delete,
                    update=update,
                )
            result = func(phylogeny_df, *args, **kwargs)
            if drop_topological_sensitivity:
                result = alifestd_drop_topological_sensitivity_polars(
                    result,
                    insert=insert,
                    delete=delete,
                    update=update,
                )
            return result

        return wrapper

    return decorator



import typing
from unittest import mock

import numpy as np
import tqdm

class Records:
    def __init__(self, init_size: int, init_root: bool = True) -> None: ...
    def addRecord(
        self,
        data_id: int,
        id: int,
        ancestor_id: int,
        search_ancestor_id: int,
        search_first_child_id: int,
        search_prev_sibling_id: int,
        search_next_sibling_id: int,
        rank: int,
        differentia: int,
    ) -> None: ...
    def mockRecord(
        self,
        data_id: int,
        id: int,
        ancestor_id: int,
        search_ancestor_id: int,
        search_first_child_id: int,
        search_prev_sibling_id: int,
        search_next_sibling_id: int,
        rank: int,
        differentia: int,
    ) -> None: ...
    def __len__(self) -> int: ...

placeholder_value: int

def collapse_unifurcations(
    records: Records, dropped_only: bool
) -> Records: ...
def copy_records_to_dict(records: Records) -> dict[str, np.ndarray]: ...
def extract_records_to_dict(records: Records) -> dict[str, np.ndarray]: ...
def extend_tree_searchtable_cpp_from_exploded(
    records: Records,
    data_ids: np.ndarray,
    num_strata_depositeds: np.ndarray,
    ranks: np.ndarray,
    differentiae: np.ndarray,
    tqdm_progress_bar: typing.Union[typing.Type[tqdm.tqdm], mock.Mock],
) -> None: ...
def build_tree_searchtable_cpp_from_exploded(
    data_ids: np.ndarray,
    num_strata_depositeds: np.ndarray,
    ranks: np.ndarray,
    differentiae: np.ndarray,
    tqdm_progress_bar: typing.Union[typing.Type[tqdm.tqdm], mock.Mock],
) -> dict[str, np.ndarray]: ...
def build_tree_searchtable_cpp_from_nested(
    data_ids: typing.List[int],
    num_strata_depositeds: typing.List[int],
    ranks: typing.List[typing.List[int]],
    differentiae: typing.List[typing.List[int]],
    tqdm_progress_bar: typing.Union[typing.Type[tqdm.tqdm], mock.Mock],
) -> dict[str, np.ndarray]: ...
def check_trie_invariant_contiguous_ids(records: Records) -> bool: ...
def check_trie_invariant_topologically_sorted(records: Records) -> bool: ...
def check_trie_invariant_chronologically_sorted(
    records: Records,
) -> bool: ...
def check_trie_invariant_single_root(records: Records) -> bool: ...
def check_trie_invariant_search_children_valid(
    records: Records,
) -> bool: ...
def check_trie_invariant_search_children_sorted(
    records: Records,
) -> bool: ...
def check_trie_invariant_no_indistinguishable_nodes(
    records: Records,
) -> bool: ...
def check_trie_invariant_data_nodes_are_leaves(
    records: Records,
) -> bool: ...
def check_trie_invariant_search_lineage_compatible(
    records: Records,
) -> bool: ...
def check_trie_invariant_ancestor_bounds(records: Records) -> bool: ...
def check_trie_invariant_root_at_zero(records: Records) -> bool: ...
def check_trie_invariant_ranks_nonnegative(
    records: Records,
) -> bool: ...
def describe_records(records: Records) -> str: ...
def diagnose_trie_invariant_contiguous_ids(records: Records) -> str: ...
def diagnose_trie_invariant_topologically_sorted(
    records: Records,
) -> str: ...
def diagnose_trie_invariant_chronologically_sorted(
    records: Records,
) -> str: ...
def diagnose_trie_invariant_single_root(records: Records) -> str: ...
def diagnose_trie_invariant_search_children_valid(
    records: Records,
) -> str: ...
def diagnose_trie_invariant_search_children_sorted(
    records: Records,
) -> str: ...
def diagnose_trie_invariant_no_indistinguishable_nodes(
    records: Records,
) -> str: ...
def diagnose_trie_invariant_data_nodes_are_leaves(
    records: Records,
) -> str: ...
def diagnose_trie_invariant_search_lineage_compatible(
    records: Records,
) -> str: ...
def diagnose_trie_invariant_ancestor_bounds(records: Records) -> str: ...
def diagnose_trie_invariant_root_at_zero(records: Records) -> str: ...
def diagnose_trie_invariant_ranks_nonnegative(
    records: Records,
) -> str: ...



import typing

import opytional as opyt

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ...genome_instrumentation import (
    HereditaryStratigraphicColumn,
    HereditaryStratumOrderedStoreList,
)
from ._calc_rank_of_first_retained_disparity_between_bsearch import (
    calc_rank_of_first_retained_disparity_between_bsearch,
)
from ._calc_rank_of_first_retained_disparity_between_generic import (
    calc_rank_of_first_retained_disparity_between_generic,
)


def calc_rank_of_first_retained_disparity_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    confidence_level: float,
) -> typing.Optional[int]:
    """Find first mismatching strata between columns.

    Implementation detail. Provides optimized implementation for special
    case where both self and second use the perfect resolution stratum
    retention policy.
    """

    impl = (
        calc_rank_of_first_retained_disparity_between_generic
        if (
            not isinstance(first, HereditaryStratigraphicColumn)
            or not isinstance(second, HereditaryStratigraphicColumn)
            or first.HasDiscardedStrata()
            or second.HasDiscardedStrata()
            # for performance reasons
            # only apply binary search to stores that support random access
            or not hasattr(first, "_stratum_ordered_store")
            or not isinstance(
                first._stratum_ordered_store, HereditaryStratumOrderedStoreList
            )
            or not hasattr(second, "_stratum_ordered_store")
            or not isinstance(
                second._stratum_ordered_store,
                HereditaryStratumOrderedStoreList,
            )
            # binary search currently requires no spurious collisions
            or first.GetStratumDifferentiaBitWidth() < 64
        )
        else calc_rank_of_first_retained_disparity_between_bsearch
    )

    res = impl(first, second, confidence_level=confidence_level)
    res = opyt.apply_if(res, lambda x: max(x, 0))  # handle negative rank case
    return res



import typing

from deprecated.sphinx import deprecated
import polars as pl

from ._alifestd_check_topological_sensitivity_polars import (
    alifestd_check_topological_sensitivity_polars,
)
from ._alifestd_warn_topological_sensitivity import (
    _alifestd_warn_topological_sensitivity,
)


@deprecated(
    version="1.23.0",
    reason="Use phyloframe.legacy.alifestd_warn_topological_sensitivity_polars instead.",
)
def alifestd_warn_topological_sensitivity_polars(
    phylogeny_df: typing.Union[pl.DataFrame, pl.LazyFrame],
    caller: str,
    *,
    insert: bool,
    delete: bool,
    update: bool,
) -> None:
    """Emit a warning if `phylogeny_df` contains columns that may be
    invalidated by topological operations.

    Parameters
    ----------
    phylogeny_df : polars.DataFrame or polars.LazyFrame
        The phylogeny as a dataframe in alife standard format.
    caller : str
        Name of the calling function, included in the warning message.
    insert : bool
        Whether the operation inserts new nodes.
    delete : bool
        Whether the operation deletes nodes.
    update : bool
        Whether the operation updates ancestor relationships.

    Input dataframe is not mutated by this operation.

    See Also
    --------
    alifestd_warn_topological_sensitivity :
        Pandas-based implementation.
    """
    _alifestd_warn_topological_sensitivity(
        alifestd_check_topological_sensitivity_polars(
            phylogeny_df,
            insert=insert,
            delete=delete,
            update=update,
        ),
        caller,
        insert=insert,
        delete=delete,
        update=update,
    )
